//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

// Process audio files (i.e. WAVE files) so that they can be
// imported into PrestoPlot. Audio files can have either one or
// two channels (i.e. mono or stereo) and have a variety of different
// sampling frequencies.

#include "stdafx.h"
#include "sgraph.h"
#include "color.h"
#include <mmsystem.h>
#include "AudioDialog.h"

#define		BITS_PER_SAMPLE		16

#define		LEFT_LOW	0
#define		LEFT_HIGH	1
#define		RIGHT_LOW	2
#define		RIGHT_HIGH	3
#define		WRITE_MOD	4

BOOL SGraph::ValidateMonoAudio(int plotId, int samplesSec, float duration, int volume)
{
	stackIterator<DataPoint> next(plot[plotId].data);
	DataPoint xy, next_xy;

	// Compute the exact bounds of the data set
	if(!next(xy))
		// Only a single data point
		return FALSE;

	next(next_xy);

	if(xy.x > next_xy.x){
		// Data is stored from high to low

		xy = next_xy;

		while(next(next_xy)){
			
			if(xy.x < next_xy.x)
				// Data is not monotonic in X
				return FALSE;
		}
	}
	else{
		// Data is stored from low to high

		xy = next_xy;

		while(next(next_xy)){
			
			if(xy.x > next_xy.x)
				// Data is not monotonic in X
				return FALSE;
		}
	}

	return TRUE;
}

BOOL SGraph::ValidateStereoAudio(int leftPlotId, int rightPlotId, int samplesSec, float duration, int volume)
{
	return (ValidateMonoAudio(leftPlotId, samplesSec, duration, volume) &&
		ValidateMonoAudio(rightPlotId, samplesSec, duration, volume));
}

void SGraph::ExportMonoAudio(CString &name, int plotId, int samplesSec, float duration, int volume)
{
	// handle to open output WAVE file
	HMMIO hmmioOut = NULL; 

	// chunk info. for output RIFF chunk
	MMCKINFO ckOutRIFF;  

	// info. for a chunk in output file
	MMCKINFO ckOut;

	// contents of 'fmt' chunks
	PCMWAVEFORMAT  pcmWaveFormat;

	// current status of <hmmioOut>
	MMIOINFO mmioinfoOut;

	// Set the format of the wave file
	pcmWaveFormat.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmWaveFormat.wBitsPerSample = BITS_PER_SAMPLE; // 16 bit for now -- make this a parameter
	pcmWaveFormat.wf.nChannels = 1; // mono
	pcmWaveFormat.wf.nSamplesPerSec = samplesSec;
	pcmWaveFormat.wf.nAvgBytesPerSec = ((pcmWaveFormat.wBitsPerSample)/8) *
                               pcmWaveFormat.wf.nChannels *
                               pcmWaveFormat.wf.nSamplesPerSec;
	pcmWaveFormat.wf.nBlockAlign = ((pcmWaveFormat.wBitsPerSample)/8) *
                               pcmWaveFormat.wf.nChannels;

	// Build the buffer. Note that when storing audio data in a 16 bit format (i.e.
	// two bytes per value per channel) the bufferSize must always be a multiple of two.
	// To insure bufferSize is a multiple of two, always round down to the nearest 
	// multiple of two.
	int bufferSize = (int)(pcmWaveFormat.wf.nAvgBytesPerSec*duration) & 0xfffffffe;
	char *audioBuffer = BuildAudioBuffer(plotId, bufferSize, samplesSec, duration, volume);

	if(audioBuffer == NULL){
		throw Error("Unable to allocate memory for audio buffer!", 
			__FILE__, __LINE__);
	}

	hmmioOut = mmioOpen((char*)(LPCSTR(name)), NULL,
                  MMIO_ALLOCBUF | MMIO_WRITE | MMIO_CREATE);

   if(hmmioOut == NULL)
      throw Error("Unable to open file", __FILE__, __LINE__);

	// Create the output file RIFF chunk of form type 'WAVE'.
	ckOutRIFF.fccType = mmioFOURCC('W', 'A', 'V', 'E');

	if(mmioCreateChunk(hmmioOut, &ckOutRIFF, MMIO_CREATERIFF) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);
		
		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// We are now descended into the 'RIFF' chunk we just created.
    // Now create the 'fmt ' chunk. Since we know the size of this chunk,
    // specify it in the MMCKINFO structure so MMIO doesn't have to seek
    // back and set the chunk size after ascending from the chunk.
    //
	ckOut.ckid = mmioFOURCC('f', 'm', 't', ' ');
	ckOut.cksize = sizeof(pcmWaveFormat);  // we know the size of this ck.
	if(mmioCreateChunk(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}
	
	// Write the PCMWAVEFORMAT structure to the 'fmt ' chunk.
	if(mmioWrite(hmmioOut, (HPSTR) &pcmWaveFormat, sizeof(pcmWaveFormat))
		!= sizeof(pcmWaveFormat)){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}
	
	
	// Ascend out of the 'fmt ' chunk, back into the 'RIFF' chunk.
	if(mmioAscend(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}
	
	// Create the 'data' chunk that holds the waveform samples.
	ckOut.ckid = mmioFOURCC('d', 'a', 't', 'a');
	if (mmioCreateChunk(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// Begin direct access of the I/O buffers.
	if(mmioGetInfo(hmmioOut, &mmioinfoOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// For each input sample, compute and write the output sample.
	for(int i = 0; i < bufferSize; i++)
	{
		// If we are the end of the output file I/O buffer, flush it.
		if(mmioinfoOut.pchNext == mmioinfoOut.pchEndWrite){
			mmioinfoOut.dwFlags |= MMIO_DIRTY;
			if(mmioAdvance(hmmioOut, &mmioinfoOut, MMIO_WRITE) != 0){
				if(hmmioOut != NULL)
					mmioClose(hmmioOut, 0);

				if(audioBuffer)
					delete [] audioBuffer;

				throw Error("Unable to write file", __FILE__, __LINE__);
			}
		}

		*(mmioinfoOut.pchNext)++ = audioBuffer[i];
	}

	// We are through processing samples, end direct access of
    // the I/O buffers. Set the MMIO_DIRTY flag for the output buffer
    // to flush it.
	mmioinfoOut.dwFlags |= MMIO_DIRTY;
	if(mmioSetInfo(hmmioOut, &mmioinfoOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// Ascend the output file out of the 'data' chunk -- this will cause
    // the chunk size of the 'data' chunk to be written.
	if(mmioAscend(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);
		
		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// Ascend the output file out of the 'RIFF' chunk -- this will cause
    // the chunk size of the 'RIFF' chunk to be written.
	if(mmioAscend(hmmioOut, &ckOutRIFF, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);
		
		if(audioBuffer)
			delete [] audioBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	if(hmmioOut != NULL)
		mmioClose(hmmioOut, 0);

	if(audioBuffer)
		delete [] audioBuffer;
}

void SGraph::ExportStereoAudio(CString &name, int leftPlotId, int rightPlotId, int samplesSec, float duration, int volume)
{

	// handle to open output WAVE file
	HMMIO hmmioOut = NULL; 

	// chunk info. for output RIFF chunk
	MMCKINFO ckOutRIFF;  

	// info. for a chunk in output file
	MMCKINFO ckOut;

	// contents of 'fmt' chunks
	PCMWAVEFORMAT  pcmWaveFormat;

	// current status of <hmmioOut>
	MMIOINFO mmioinfoOut;

	// Set the format of the wave file
	pcmWaveFormat.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmWaveFormat.wBitsPerSample = BITS_PER_SAMPLE; // 16 bit for now -- make this a parameter
	pcmWaveFormat.wf.nChannels = 2; // stereo
	pcmWaveFormat.wf.nSamplesPerSec = samplesSec;
	pcmWaveFormat.wf.nAvgBytesPerSec = ((pcmWaveFormat.wBitsPerSample)/8) *
                               pcmWaveFormat.wf.nChannels *
                               pcmWaveFormat.wf.nSamplesPerSec;
	pcmWaveFormat.wf.nBlockAlign = ((pcmWaveFormat.wBitsPerSample)/8) *
                               pcmWaveFormat.wf.nChannels;

	// Build the buffer. Note that when storing audio data in a 16 bit format (i.e.
	// two bytes per value per channel) the bufferSize must always be a multiple of two.
	// To insure bufferSize is a multiple of two, always round down to the nearest 
	// multiple of two.
	int bufferSize = (int)(samplesSec*((pcmWaveFormat.wBitsPerSample)/8)*duration) & 0xfffffffe;;
	char* leftBuffer, *rightBuffer;
	
	leftBuffer = BuildAudioBuffer(leftPlotId, bufferSize, samplesSec, duration, volume);

	if(leftBuffer == NULL){
		throw Error("Unable to allocate memory for left channel audio buffer!", 
			__FILE__, __LINE__);
	}

	rightBuffer = BuildAudioBuffer(rightPlotId, bufferSize, samplesSec, duration, volume);

	if(rightBuffer == NULL){
		throw Error("Unable to allocate memory for right channel audio buffer!", 
			__FILE__, __LINE__);
	}

	hmmioOut = mmioOpen((char*)(LPCSTR(name)), NULL,
                  MMIO_ALLOCBUF | MMIO_WRITE | MMIO_CREATE);

   if(hmmioOut == NULL)
      throw Error("Unable to open file", __FILE__, __LINE__);

	// Create the output file RIFF chunk of form type 'WAVE'.
	ckOutRIFF.fccType = mmioFOURCC('W', 'A', 'V', 'E');

	if(mmioCreateChunk(hmmioOut, &ckOutRIFF, MMIO_CREATERIFF) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);
		
		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// We are now descended into the 'RIFF' chunk we just created.
    // Now create the 'fmt ' chunk. Since we know the size of this chunk,
    // specify it in the MMCKINFO structure so MMIO doesn't have to seek
    // back and set the chunk size after ascending from the chunk.
    //
	ckOut.ckid = mmioFOURCC('f', 'm', 't', ' ');
	ckOut.cksize = sizeof(pcmWaveFormat);  // we know the size of this ck.
	if(mmioCreateChunk(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}
	
	// Write the PCMWAVEFORMAT structure to the 'fmt ' chunk.
	if(mmioWrite(hmmioOut, (HPSTR) &pcmWaveFormat, sizeof(pcmWaveFormat))
		!= sizeof(pcmWaveFormat)){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}
	
	
	// Ascend out of the 'fmt ' chunk, back into the 'RIFF' chunk.
	if(mmioAscend(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}
	
	// Create the 'data' chunk that holds the waveform samples.
	ckOut.ckid = mmioFOURCC('d', 'a', 't', 'a');
	if (mmioCreateChunk(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// Begin direct access of the I/O buffers.
	if(mmioGetInfo(hmmioOut, &mmioinfoOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// For each input sample, compute and write the output sample.
	for(int i = 0; i < bufferSize; i += 2 /* write both left and right channels */)
	{
		// If we are the end of the output file I/O buffer, flush it.
		if(mmioinfoOut.pchNext == mmioinfoOut.pchEndWrite){
			mmioinfoOut.dwFlags |= MMIO_DIRTY;
			if(mmioAdvance(hmmioOut, &mmioinfoOut, MMIO_WRITE) != 0){
				if(hmmioOut != NULL)
					mmioClose(hmmioOut, 0);

				if(leftBuffer)
					delete [] leftBuffer;

				if(rightBuffer)
					delete [] rightBuffer;

				throw Error("Unable to write file", __FILE__, __LINE__);
			}
		}

		// Pack the data as follows (LOB = "Low Order Byte" and HOB = "High Order Byte"):
		// | left channel LOB | left channel HOB | right channel LOB | right channel HOB |
		*(mmioinfoOut.pchNext)++ = leftBuffer[i];
		*(mmioinfoOut.pchNext)++ = leftBuffer[i + 1];

		// If we are the end of the output file I/O buffer, flush it.
		if(mmioinfoOut.pchNext == mmioinfoOut.pchEndWrite){
			mmioinfoOut.dwFlags |= MMIO_DIRTY;
			if(mmioAdvance(hmmioOut, &mmioinfoOut, MMIO_WRITE) != 0){
				if(hmmioOut != NULL)
					mmioClose(hmmioOut, 0);

				if(leftBuffer)
					delete [] leftBuffer;

				if(rightBuffer)
					delete [] rightBuffer;

				throw Error("Unable to write file", __FILE__, __LINE__);
			}
		}

		*(mmioinfoOut.pchNext)++ = rightBuffer[i];
		*(mmioinfoOut.pchNext)++ = rightBuffer[i + 1];
	}

	// We are through processing samples, end direct access of
    // the I/O buffers. Set the MMIO_DIRTY flag for the output buffer
    // to flush it.
	mmioinfoOut.dwFlags |= MMIO_DIRTY;
	if(mmioSetInfo(hmmioOut, &mmioinfoOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);

		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// Ascend the output file out of the 'data' chunk -- this will cause
    // the chunk size of the 'data' chunk to be written.
	if(mmioAscend(hmmioOut, &ckOut, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);
		
		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	// Ascend the output file out of the 'RIFF' chunk -- this will cause
    // the chunk size of the 'RIFF' chunk to be written.
	if(mmioAscend(hmmioOut, &ckOutRIFF, 0) != 0){
		if(hmmioOut != NULL)
			mmioClose(hmmioOut, 0);
		
		if(leftBuffer)
			delete [] leftBuffer;

		if(rightBuffer)
			delete [] rightBuffer;

		throw Error("Unable to write file", __FILE__, __LINE__);
	}

	if(hmmioOut != NULL)
		mmioClose(hmmioOut, 0);

	if(leftBuffer)
		delete [] leftBuffer;

	if(rightBuffer)
		delete [] rightBuffer;
}

char* SGraph::BuildAudioBuffer(int plotId, int bufferSize, int samplesSec, float duration, int volume)
{
	BOOL reverse = FALSE;
	DataPoint xy, xy_next;
	char* buffer = new char [bufferSize];
	char* tail; 
	char* ptr;

	if(buffer == NULL)
		return NULL;

	tail = buffer + bufferSize - 1;

	stackIterator<DataPoint> next(plot[plotId].data);
	
	// Compute the exact bounds of the data set
	next(xy);

	float audioMinX = xy.x, 
		  audioMaxX = xy.x, 
		  audioMinY = xy.y, 
		  audioMaxY = xy.y;

	while(next(xy)){
		if(audioMinX > xy.x)
			audioMinX = xy.x;

		if(audioMaxX < xy.x)
			audioMaxX = xy.x;

		if(audioMinY > xy.y)
			audioMinY = xy.y;

		if(audioMaxY < xy.y)
			audioMaxY = xy.y;
	}

	next.reset(plot[plotId].data);

	// Do we need to pack in reverse order?
	next(xy);
	next(xy_next);

	if(xy.x > xy_next.x)
		reverse = TRUE;

	// Uniformly sample the plot in intervals of dx. Note that the
	// product of samplesSec and duration is always truncated to an
	// integer.
	double dx = (audioMaxX - audioMinX)/(int(samplesSec*duration) - 1.0);
	double dy = audioMaxY - audioMinY;
	double y_offset = audioMinY;
	double curr_x = xy.x;
	int ADY;

	// Pack the buffer by interpolating and scaling the plot data.
	// For now, assume that we will only write 16 bit data.
	if(reverse){
		ptr = tail;

		while(curr_x >= xy_next.x){
			double y = xy.y + ((curr_x - xy.x)/(xy_next.x - xy.x))*(xy_next.y - xy.y);
			
			// Digitize y
			ADY = int((y - y_offset)*volume*65536/(dy*100)) - 32768;

			// Store the low order byte
			*(ptr - 1) = ADY & 0x000000ff;

			// Store the high order byte
			*(ptr) = (ADY & 0x0000ff00) >> 8;
			
			ptr -= 2;

			if(ptr == buffer)
				return buffer;
			
			curr_x -= dx;
		}
		
		xy = xy_next;

		while(next(xy_next)){

			while(curr_x >= xy_next.x){
				double y = xy.y + ((curr_x - xy.x)/(xy_next.x - xy.x))*(xy_next.y - xy.y);
				
				// Digitize y
				ADY = int((y - y_offset)*volume*65536/(dy*100)) - 32768;
				
				// Store the low order byte
				*(ptr - 1) = ADY & 0x000000ff;

				// Store the high order byte
				*(ptr) = (ADY & 0x0000ff00) >> 8;
				
				ptr -= 2;
				
				if(ptr == buffer)
					return buffer;

				curr_x -= dx;
			}

			xy = xy_next;
		}

		// Pad the remaining buffer (a possibility due to round off error)
		while(ptr > buffer){
			ADY = 32768;
				
			// Store the low order byte
			*(ptr - 1) = ADY & 0x000000ff;

			// Store the high order byte
			*(ptr) = (ADY & 0x0000ff00) >> 8;
			
			ptr -= 2;
		}

		return buffer;

	}
	else{
		ptr = buffer;

		while(curr_x <= xy_next.x){
			double y = xy.y + ((curr_x - xy.x)/(xy_next.x - xy.x))*(xy_next.y - xy.y);
			
			// Digitize y
			ADY = int((y - y_offset)*volume*65536/(dy*100)) - 32768;
			
			// Store the low order byte
			*(ptr++) = ADY & 0x000000ff;

			// Store the high order byte
			*(ptr++) = (ADY & 0x0000ff00) >> 8;
			
			if(ptr == tail)
				return buffer;

			curr_x += dx;
		}
		
		xy = xy_next;

		while(next(xy_next)){

			while(curr_x <= xy_next.x){
				double y = xy.y + ((curr_x - xy.x)/(xy_next.x - xy.x))*(xy_next.y - xy.y);
				
				// Digitize y
				ADY = int((y - y_offset)*volume*65536/(dy*100)) - 32768;
				
				// Store the low order byte
				*(ptr++) = ADY & 0x000000ff;

				// Store the high order byte
				*(ptr++) = (ADY & 0x0000ff00) >> 8;
				
				if(ptr == tail)
					return buffer;

				curr_x += dx;
			}

			xy = xy_next;
		}

		// Pad the remaining buffer (a possibility due to round off error)
		while(ptr < tail){
			ADY = 32768;

			// Store the low order byte
			*(ptr++) = ADY & 0x000000ff;

			// Store the high order byte
			*(ptr++) = (ADY & 0x0000ff00) >> 8;
		}

		return buffer;
	}

	// We should never get here!
	return NULL;
}