//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	The IOBuffer provides a general mechanism to handle both file and
//	memory input for PrestoPlot. File input will typically be initiated
//	by the user providing a file name, while memory input will arise from
//	accessing the Windows clipboard.

#ifndef __IO_BUFFER
#define __IO_BUFFER

#include <fstream.h>
#include "excepts.h"

class IOBuffer {
private:
	fstream *file;
	HANDLE mem;
	int bufType;
public:
	IOBuffer(fstream &m_file)
	{
		file = &m_file;
		mem = NULL;
		
		bufType = IO_FILE;
	};

	IOBuffer(HANDLE &m_mem)
	{
		file = NULL;
		mem = m_mem;
		
		bufType = IO_MEM;
	};

	// Enumerate the possible IO buffer types
	enum {IO_FILE = 0, IO_MEM};

	// Are we at the end of the io buffer?
	inline BOOL eof()
	{
		switch(bufType){
			case IO_FILE:
				if(file == NULL)
					throw Error("NULL file pointer!", 
						__FILE__, __LINE__);

				return file->eof();
				break;
			case IO_MEM:
				if(mem == NULL)
					throw Error("NULL mem pointer!", 
						__FILE__, __LINE__);

				return (((char*)mem)[0] == '\0');
				break;
		};

		// Shouldn't get here!
		return TRUE;
	};

	// Read n characters from the io buffer and place them into 
	// buffer
	inline void getline(char *buffer, int n)
	{
		switch(bufType){
			case IO_FILE:
				if(file == NULL)
					throw Error("NULL file pointer!", __FILE__, __LINE__);

				file->getline(buffer, n);
				break;
			case IO_MEM:
				{
					if(n == 0)
						return;

					if(mem == NULL)
						throw Error("NULL mem pointer!", __FILE__, __LINE__);

					int len;

					// The size of the offset depends on the kind of
					// data we're reading. offset = 2 if lines are terminated
					// by "\r\n", and offset = 1 if lines are terminated
					// by "\n".
					int offset = 2;

					// Find the first instance of '\r' in mem
					char *ptr = strchr((char *)mem, '\r');
					
					if(ptr == NULL){
						// If we couldn't find '\n', are we dealing with a
						// unix style character format?
						ptr = strchr((char *)mem, '\n');

						offset = 1;
					}


					if(ptr == NULL){
						// Grab the entire string
						len = strlen((char *)mem);

						// We are reading a line that is not terminated
						// by '\r' or '\n'
						offset = 0;
					}
					else{
						//len = strlen((char *)mem) - strlen(ptr);
						len = ptr - (char *)mem;
					}

					len = min(len, n);

					strncpy(buffer, (char *)mem, len);
					
					// Terminate the string
					buffer[len] = '\0';

					// Increment the pointer to the memory buffer.
					mem = (char*)mem + len + offset;
				}

				break;
		};
	};
};


#endif
