//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

// Process audio files (i.e. WAVE files) so that they can be
// imported into PrestoPlot. Audio files can have either one or
// two channels (i.e. mono or stereo) and have a variety of different
// sampling frequencies.

#include "stdafx.h"
#include "sgraph.h"
#include "color.h"
#include <mmsystem.h>

int SGraph::ImportAudioPlot(CString &name, int type)
{
	// Only wave files are currently supported!
	switch(type){
		case AUDIO_WAVE:
			break;
		default:
			AfxMessageBox("Only WAVE files are currently supported!");
			return S_ERROR_IO;
	};
	
	//////////////////////////////////////////////////////////////
	// WAVE file processing based on MS documentation (lowpass.c).
	//////////////////////////////////////////////////////////////

	// handle to open input WAVE file
	HMMIO hmmioIn = NULL;

	// chunk info. for input RIFF chunk
	MMCKINFO ckInRIFF;

	// info. for a chunk in input file
	MMCKINFO ckIn;

	// contents of 'fmt' chunks
	PCMWAVEFORMAT pcmWaveFormat; 

	// current status of <hmmioIn>
	MMIOINFO mmioinfoIn;

	// Open the input file for reading using buffered I/O.
	hmmioIn = mmioOpen((char*)LPCSTR(name), NULL, MMIO_ALLOCBUF | MMIO_READ);

	if(hmmioIn == NULL)
		return S_ERROR_IO_OPEN;  // cannot open the WAVE file

	// Descend the input file into the 'RIFF' chunk.
	if(mmioDescend(hmmioIn, &ckInRIFF, NULL, 0) != 0){
		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);
		
		return S_ERROR_AUDIO;
	}

	// Make sure the input file is a WAVE file.
	if((ckInRIFF.ckid != FOURCC_RIFF) ||
	   (ckInRIFF.fccType != mmioFOURCC('W', 'A', 'V', 'E'))){

		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		AfxMessageBox("Unable to locate the 'RIFF' chunk - are you sure"
			"this is a WAV file?");

		return S_ERROR_AUDIO;
	}

	// Search the input file for for the 'fmt ' chunk.
	ckIn.ckid = mmioFOURCC('f', 'm', 't', ' ');
	if(mmioDescend(hmmioIn, &ckIn, &ckInRIFF, MMIO_FINDCHUNK) != 0){
		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		AfxMessageBox("Unable to locate the 'fmt' chunk");

		return S_ERROR_AUDIO;
	}

	// Expect the 'fmt' chunk to be at least as large as <pcmWaveFormat>;
	// if there are extra parameters at the end, we'll ignore them
	if(ckIn.cksize < (long) sizeof(pcmWaveFormat)){
		// 'fmt ' chunk too small

		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		AfxMessageBox("Unable to locate the 'fmt' chunk");

		return S_ERROR_AUDIO;
	}
	
	// Read the 'fmt ' chunk into <pcmWaveFormat>.
	if(mmioRead(hmmioIn, (HPSTR) &pcmWaveFormat,
		(long) sizeof(pcmWaveFormat)) != (long) sizeof(pcmWaveFormat)){
		// truncated file, probably
		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		return S_ERROR_AUDIO;
	}
	
	// Ascend the input file out of the 'fmt ' chunk.
	if(mmioAscend(hmmioIn, &ckIn, 0) != 0){
		// truncated file, probably
		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		return S_ERROR_AUDIO;
	}

	// Make sure the input file is a PCM WAVE file.
   if((pcmWaveFormat.wf.wFormatTag != WAVE_FORMAT_PCM)){
		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		return S_ERROR_AUDIO;
	}

   // Search the input file for for the 'data' chunk.
   ckIn.ckid = mmioFOURCC('d', 'a', 't', 'a');
   if (mmioDescend(hmmioIn, &ckIn, &ckInRIFF, MMIO_FINDCHUNK) != 0){
		// no 'data' chunk
		if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		return S_ERROR_AUDIO;
	}  

   // Begin direct access of the I/O buffers.
   if (mmioGetInfo(hmmioIn, &mmioinfoIn, 0) != 0){
	   if(hmmioIn != NULL)
			mmioClose(hmmioIn, 0);

		return S_ERROR_AUDIO;
   }

	BOOL isStero = (pcmWaveFormat.wf.nChannels == 2);
	int audioSize = pcmWaveFormat.wBitsPerSample/8;
	int bytesToRead = pcmWaveFormat.wf.nChannels * audioSize;
	long dataSize = ckIn.cksize / bytesToRead;

	// The time step is equal to the inverse frequency and has units
	// of sec.
	float dt = 1.0f/pcmWaveFormat.wf.nSamplesPerSec; 

	DataPoint xy;

	if(isStero){
		SPlot leftPlot, rightPlot;

		// Assume data oscillates about zero
		leftPlot.canXLog = TRUE;
		leftPlot.canYLog = FALSE;
					
		// Save the file name
		leftPlot.info.Format("Left channel: %s", name);

		// Assume data oscillates about zero
		rightPlot.canXLog = TRUE;
		rightPlot.canYLog = FALSE;
					
		// Save the file name
		rightPlot.info.Format("Right channel: %s", name);

		try{
			plot.Push(leftPlot);
			plot.Push(rightPlot);
		}
		catch(...){
			// Memory error!
			return S_ERROR_MEM_ALLOC;
		}

		// Initialize the plot bounding box
		SetBoundingBox();

		int leftPlotId = plot.Length() - 2;
		int rightPlotId = leftPlotId + 1;

		// Increment the line colors
		if(plot.Length() == 2){
			plot[leftPlotId].lineColor = IndexToColor(S_CLR_BLACK);
			plot[rightPlotId].lineColor = IndexToColor(S_CLR_RED);
		}
		else{
			plot[leftPlotId].lineColor = 
					IndexToColor((ColorToIndex(plot[leftPlotId - 1].lineColor) + 1) % 
					S_CLR_NONE);
				
			// Exclude white from the list of colors!
			if(ColorToIndex(plot[leftPlotId].lineColor) == S_CLR_WHITE)
				plot[leftPlotId].lineColor = IndexToColor(S_CLR_WHITE + 1);

			plot[rightPlotId].lineColor = 
					IndexToColor((ColorToIndex(plot[rightPlotId - 1].lineColor) + 1) % 
					S_CLR_NONE);
				
			// Exclude white from the list of colors!
			if(ColorToIndex(plot[rightPlotId].lineColor) == S_CLR_WHITE)
				plot[rightPlotId].lineColor = IndexToColor(S_CLR_WHITE + 1);
		}

		for(long numSamples = 1;numSamples <= dataSize; numSamples ++){

			// If we are at the end of the input file I/O buffer, fill it.
			// Test to see that we don't hit end of file.
			if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead)
			{
				if(mmioAdvance(hmmioIn, &mmioinfoIn, MMIO_READ) != 0){
					if(hmmioIn != NULL)
						mmioClose(hmmioIn, 0);

					return S_ERROR_AUDIO;
				}
				
				if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead){
					if(hmmioIn != NULL)
						mmioClose(hmmioIn, 0);

					return S_ERROR_AUDIO;
				}
			}

			// Read in two audio channels and convert both to floats.
			int left = 0, right = 0;
			
			// The easiest way to convert signed values
			switch(audioSize){
				case 1:
					left = *((char*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext++;
					break;
				case 2:
					left = *((short*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext += 2;
					break;
				case 4:
					left = *((int*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext += 4;
					break;
				default:
					return S_ERROR_AUDIO;
			};

			// If we are at the end of the input file I/O buffer, fill it.
			// Test to see that we don't hit end of file.
			if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead)
			{
				if(mmioAdvance(hmmioIn, &mmioinfoIn, MMIO_READ) != 0){
					if(hmmioIn != NULL)
						mmioClose(hmmioIn, 0);

					return S_ERROR_AUDIO;
				}
				
				if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead){
					if(hmmioIn != NULL)
						mmioClose(hmmioIn, 0);

					return S_ERROR_AUDIO;
				}
			}

			// The easiest way to convert signed values
			switch(audioSize){
				case 1:
					right = *((char*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext++;
					break;
				case 2:
					right = *((short*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext += 2;
					break;
				case 4:
					right = *((int*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext += 4;
					break;
				default:
					return S_ERROR_AUDIO;
			};
			
			xy.x = numSamples*dt;
			xy.y = (float)(left);
			
			plot[leftPlotId].data.Push(xy);

			xy.y = (float)(right);
			
			plot[rightPlotId].data.Push(xy);
		}
	}
	else{ // isMono
		SPlot monoPlot;

		// Assume data oscillates about zero
		monoPlot.canXLog = TRUE;
		monoPlot.canYLog = FALSE;
					
		// Save the file name
		monoPlot.info.Format("Mono channel: %s", name);

		try{
			plot.Push(monoPlot);
		}
		catch(...){
			// Memory error!
			return S_ERROR_MEM_ALLOC;
		}
		
		// Initialize the plot bounding box
		SetBoundingBox();

		int monoPlotId = plot.Length() - 1;

		// Increment the line color if there is already an existing plot
		if(plot.Length() != 1){
			plot[monoPlotId].lineColor = 
					IndexToColor((ColorToIndex(plot[monoPlotId - 1].lineColor) + 1) % 
					S_CLR_NONE);
				
			// Exclude white from the list of colors!
			if(ColorToIndex(plot[monoPlotId].lineColor) == S_CLR_WHITE)
				plot[monoPlotId].lineColor = IndexToColor(S_CLR_WHITE + 1);
		}

		for(long numSamples = 1;numSamples <= dataSize; numSamples ++){
			// If we are at the end of the input file I/O buffer, fill it.
			// Test to see that we don't hit end of file while (lSamples > 0).
			if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead)
			{
				if(mmioAdvance(hmmioIn, &mmioinfoIn, MMIO_READ) != 0){
					if(hmmioIn != NULL)
						mmioClose(hmmioIn, 0);

					return S_ERROR_AUDIO;
				}
				
				if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead){
					if(hmmioIn != NULL)
						mmioClose(hmmioIn, 0);

					return S_ERROR_AUDIO;
				}
			}

			// Read in a single audio channels and convert it to a float.
			int mono = 0;

			// The easiest way to convert signed values
			switch(audioSize){
				case 1:
					mono = *((char*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext++;
					break;
				case 2:
					mono = *((short*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext += 2;
					break;
				case 4:
					mono = *((int*)mmioinfoIn.pchNext);
					mmioinfoIn.pchNext += 4;
					break;
				default:
					return S_ERROR_AUDIO;
			};

			xy.x = numSamples*dt;

			if(audioSize == 1){
				if(mono > 0.0)
					xy.y = (float)(mono - 128);
				else
					xy.y = (float)(mono + 128);
			}
			else
				xy.y = (float)mono;
			
			plot[monoPlotId].data.Push(xy);
		}
	}

	if(hmmioIn != NULL)
		mmioClose(hmmioIn, 0);

	// No errors!
	return -1;
}