//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

#include "stdafx.h"
#include "sgraph.h"
#include "sgraphdoc.h"
#include "sgraphview.h"
#include "ExportRasterImage.h"
#include "color.h"
#include "Excepts.h"
#include "png.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Header ID needed for writing bitmaps to disk.
#define BFT_BITMAP 0x4d42   /* 'BM' */ 

#define	PRESTO_PNG_COLOR_DEPTH	8
#define	PRESTO_BMP_COLOR_DEPTH	24

void PrestoPNGErrorCallBack(png_structp png_ptr, png_const_charp msg);

void PrestoPNGErrorCallBack(png_structp png_ptr, png_const_charp msg)
{
	throw Error(msg);
}

HBITMAP CSgraphView::MakeDIB(CDC* pDC, int dib_width, int dib_height, int dib_bpp)
{
	BITMAPINFO bmpInfoInput;
	LPVOID pBitmapBits;
	int bitmapSize;
	HBITMAP hDib;
	HGDIOBJ hOldDib;
	CDC renderDC;
	int i;

	CSgraphDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// Initialize the input bitmap header information...
	memset(&bmpInfoInput, 0, sizeof(BITMAPINFO));
	bmpInfoInput.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	
	bmpInfoInput.bmiHeader.biWidth = dib_width;
	bmpInfoInput.bmiHeader.biHeight = dib_height;
	bmpInfoInput.bmiHeader.biPlanes = 1;
	bmpInfoInput.bmiHeader.biBitCount = dib_bpp;
	bmpInfoInput.bmiHeader.biCompression = BI_RGB;
	bmpInfoInput.bmiHeader.biSizeImage = bmpInfoInput.bmiHeader.biWidth * 
		bmpInfoInput.bmiHeader.biHeight * (bmpInfoInput.bmiHeader.biBitCount/8);

	// Create a DIB in memory that we can render to.
	hDib = CreateDIBSection(pDC->GetSafeHdc(), &bmpInfoInput, 
		DIB_RGB_COLORS, &pBitmapBits, NULL, (DWORD)0);

	// Save the image size in bytes
	bitmapSize = bmpInfoInput.bmiHeader.biSizeImage;

	// Create memory device context
	renderDC.CreateCompatibleDC(pDC);

	hOldDib = renderDC.SelectObject(hDib);

	// Force the OnDraw() function to render to our DIB (regardless of the
	// state of the double buffering).
	isPrinting = TRUE;

	int tmp_width, tmp_height;

	// Save the current viewport dimensions
	tmp_width = extent.Width();
	tmp_height = extent.Height();

	extent = CRect(0, 0, dib_width, dib_height);

	// Draw a background color
	// If the current background color is NO_COLOR, 
	// use the system window color.
	CBrush bkgrnd(pDoc->background_color == NO_COLOR ? 
		GetSysColor(COLOR_WINDOW): pDoc->background_color);

	renderDC.FillRect(extent, &bkgrnd);

	// Draw to the device context.
	// The isPrinting flag disables the graph selection boxes
	for(i = 0;i < pDoc->annotate_text.Length();i++)
		pDoc->annotate_text[i].Printing(TRUE);

	for(i = 0;i < pDoc->num_graph;i++)
		pDoc->graph[i].Printing(TRUE);

	OnDraw(&renderDC);

	for(i = 0;i < pDoc->annotate_text.Length();i++)
		pDoc->annotate_text[i].Printing(FALSE);

	for(i = 0;i < pDoc->num_graph;i++)
		pDoc->graph[i].Printing(FALSE);

	// Restore the original viewport dimensions
	extent = CRect(0, 0, tmp_width, tmp_height);

	// Restore the "is printing" flag
	isPrinting = FALSE;

	if(hOldDib){
		renderDC.SelectObject(hOldDib);
		hOldDib = NULL;
	}

	return hDib;
}

void CSgraphView::ExportImageBMP(const char* filename)
{
	// Get the requested raster resolution
	CExportRasterImage eriDlg(extent.Height(), extent.Width());

	if(eriDlg.DoModal() != IDOK)
		return;

	ExportImageBMP(filename, eriDlg.m_width, eriDlg.m_height);
}

void CSgraphView::ExportImageBMP(const char* filename, int width, int height)
{
	FILE *fout;

	fout = fopen(filename, "wb");

	if(!fout){
		AfxMessageBox("Unable to open " + CString(filename) + " for writing");
		return;
	}

	// Set the hourglass cursor - this may take a while
	::SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Get the current (screen) device context
	CDC *pDC = GetDC();

	HBITMAP hDib = MakeDIB(pDC, width, height, PRESTO_BMP_COLOR_DEPTH);

	if(hDib == NULL){
		AfxMessageBox("Unable to write bitmap (invalid dib)");
		fclose(fout);
		return;
	}

	DIBSECTION dibInfo;
	GetObject(hDib, sizeof(DIBSECTION), &dibInfo);

	BITMAPFILEHEADER hdr;

	// Fill in the fields of the file header
	hdr.bfType = BFT_BITMAP; 
	hdr.bfSize = dibInfo.dsBmih.biSizeImage + sizeof(BITMAPINFOHEADER) +
		sizeof (BITMAPFILEHEADER);
	hdr.bfReserved1 = 0;
	hdr.bfReserved2 = 0;
	hdr.bfOffBits   = (DWORD)sizeof(BITMAPFILEHEADER);

	// Write the file header
	fwrite(&hdr, sizeof(BITMAPFILEHEADER), 1, fout);
	
	// Write the bitmap header
	fwrite(&(dibInfo.dsBmih), sizeof(BITMAPINFOHEADER), 1, fout);

	// Write the body of the bitmap data
	fwrite(dibInfo.dsBm.bmBits, dibInfo.dsBmih.biSizeImage, 1, fout);

	if(fout)
		fclose(fout);

	::SetCursor(LoadCursor(NULL, IDC_ARROW));	

	ReleaseDC(pDC);

	DeleteObject(hDib);

}

void CSgraphView::ExportImagePNG(const char* filename)
{
	// Get the requested raster resolution
	CExportRasterImage eriDlg(extent.Height(), extent.Width());

	if(eriDlg.DoModal() != IDOK)
		return;

	ExportImagePNG(filename, eriDlg.m_width, eriDlg.m_height);
}

void CSgraphView::ExportImagePNG(const char* filename, int width, int height)
{
	FILE *fout;

	fout = fopen(filename, "wb");

	if(!fout){
		AfxMessageBox("Unable to open " + CString(filename) + " for writing");
		return;
	}

	// Set the hourglass cursor - this may take a while
	::SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Get the current (screen) device context
	CDC *pDC = GetDC();

	// *************************************************************************************
	// The PRESTO_PNG_COLOR_DEPTH specifies the bits per COLOR. MakeDIB expect bits-per-pixel.
	// Hence the 3*PRESTO_PNG_COLOR_DEPTH below.
	// *************************************************************************************
	HBITMAP hDib = MakeDIB(pDC, width, height, 3*PRESTO_PNG_COLOR_DEPTH);

	if(hDib == NULL){
		AfxMessageBox("Unable to write bitmap (invalid dib)");
		fclose(fout);
		return;
	}

	DIBSECTION dibInfo;
	GetObject(hDib, sizeof(DIBSECTION), &dibInfo);

	// Now that we have a DIB, the real fun begins! We need to write
	// the raw data in the png format. Based on the example code
	// included with the libpng distribution.
	png_structp png_ptr;
	png_infop info_ptr;
	png_text text_info[2];
	png_byte *row_ptr = NULL;

	try{
		
		png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
			NULL, PrestoPNGErrorCallBack, NULL);

		if(png_ptr == NULL){
			throw Error("Unable to initialize PNG data");
		}
		
		info_ptr = png_create_info_struct(png_ptr);

		if(info_ptr == NULL){
			throw Error("Unable to write png file (png_create_info_struct)");
		}

		// Set up the output control using standard C streams
		png_init_io(png_ptr, fout);

		png_set_IHDR(png_ptr, info_ptr, dibInfo.dsBm.bmWidth, dibInfo.dsBm.bmHeight, 
			PRESTO_PNG_COLOR_DEPTH, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, 
			PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

		// Write comments into the image
		text_info[0].key = "Title";
		text_info[0].text = "A Graph Image";
		text_info[0].compression = PNG_TEXT_COMPRESSION_NONE;
		text_info[1].key = "Author";
		text_info[1].text = "PrestoPlot";
		text_info[1].compression = PNG_TEXT_COMPRESSION_NONE;

		png_set_text(png_ptr, info_ptr, text_info, 2);

		// Write the file header information.
		png_write_info(png_ptr, info_ptr);

		// Write the dib one row at a time
		int row_size = dibInfo.dsBmih.biSizeImage/dibInfo.dsBmih.biHeight;

		// The bytes per pixel for the output png file, not the input DIB file
		int bytes_per_pixel = 3;

		unsigned char *row_ptr = new unsigned char [bytes_per_pixel*dibInfo.dsBmih.biWidth];
		unsigned char *dib_buffer = (unsigned char *)(dibInfo.dsBm.bmBits);

		if(row_ptr == NULL)
			throw Error("Unable to write png file (insufficient memory)");

		// Flip the y dimension
		for(int i = dibInfo.dsBmih.biHeight - 1;i >= 0;i--){
			for(int j = 0;j < dibInfo.dsBmih.biWidth;j++){

				// Swap the color packing; RGB->BGR
				row_ptr[j*bytes_per_pixel]  = dib_buffer[i*row_size + j*3+2];
				row_ptr[j*bytes_per_pixel+1]= dib_buffer[i*row_size + j*3+1];
				row_ptr[j*bytes_per_pixel+2]= dib_buffer[i*row_size + j*3];
			}
			
			png_write_row(png_ptr, row_ptr);
		}

		png_write_end(png_ptr, info_ptr);
	}
	catch(Error &e){
		AfxMessageBox(e.message());
	}
	catch(...){
		AfxMessageBox("Unspecified error writing png file");
	}

	if(fout)
		fclose(fout);

	::SetCursor(LoadCursor(NULL, IDC_ARROW));	

	if(info_ptr){
		png_destroy_write_struct(&png_ptr, &info_ptr);
	}
	else{
		png_destroy_write_struct(&png_ptr, png_infopp_NULL);
	}
	
	if(row_ptr)
		delete [] row_ptr;

	ReleaseDC(pDC);

	DeleteObject(hDib);

}