//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
#include "stdafx.h"
#include "Text.h"
#include "Presto.h"
#include "Excepts.h"
#include <math.h>
#include "List.h"

// Token types
enum {TT_NORMAL, TT_LATEX, TT_SUPER, TT_SUB};

class TextToken{
public:
	// Is this token Latex?
	BOOL latex;
	char *token;
};

// Helper functions
BOOL GetTokenList(char *buffer, list<TextToken> &token_list);
char *MatchCurlyBrace(char *buffer);
FIXED FloatToFixed(float f);

int TextElem::printScaleMul = 1;
int TextElem::printScaleDiv = 1;

TextElem::TextElem()
{
	for(int i = TEXT_TOP;i <= TEXT_BOTTOM;i++)
		next[i] = NULL;
	
	color = 0;
	font = PRESTO_FONT_DEFAULT;
	curr_font = NULL;
	SetAngle(0);
	elem = _T("");
	options = TEXT_CENTER_THIS;

	// Zero length strings will keep these default values for
	// elem_height and offset.
	elem_height = 0;
	offset = 0;

	UpdateAll();
}

TextElem::TextElem(CString &m_text, int m_font, int m_size, int m_angle,
		COLORREF m_color, int m_opt)
{
	for(int i = TEXT_TOP;i <= TEXT_BOTTOM;i++)
		next[i] = NULL;
	
	SetAngle(m_angle);
	SetPointSize(m_size);
	color = m_color;
	font = m_font;
	curr_font = NULL;
	elem = m_text;
	options = m_opt;

	// Zero length strings will keep these default values for
	// elem_height and offset.
	elem_height = 0;
	offset = 0;

	UpdateAll();
};

TextElem::~TextElem()
{
	Flush();

	if(curr_font)
		delete curr_font;
}

void TextElem::Flush()
{
	for(int i = TEXT_TOP;i <= TEXT_BOTTOM;i++)
		if(next[i]){
			delete next[i];
			next[i] = NULL;
		}

	UpdateAll();
}

void TextElem::Copy(const TextElem &m_te)
{
	Flush();
	
	pt_size = m_te.pt_size;
	color = m_te.color;
	font = m_te.font;
	SetAngle(m_te.angle);
	elem = m_te.elem;
	options = m_te.options;
	elem_height = m_te.elem_height;
	offset = m_te.offset;

	for(int i = TEXT_TOP;i <= TEXT_BOTTOM;i++)
		if(m_te.next[i]){
			next[i] = new TextElem;
			next[i]->Copy(*(m_te.next[i]));
		}
}

CFont* TextElem::SetFont(CDC *pDC, LOGFONT *m_font_table)
{
	// CreatePointFontIndirect measures font size in 10^th of a point
	m_font_table[font].lfHeight = -10*MulDiv(pt_size, printScaleMul, printScaleDiv);
	m_font_table[font].lfEscapement = angle;
	m_font_table[font].lfOrientation = m_font_table[font].lfEscapement;

	// Save a copy of the current GDI font object (for later deletion)
	CFont *old_ptr = curr_font;
	
	curr_font = new CFont;

	// Use CreatePointFontIndirect instead of CreateFontIndirect to get
	// better consistancy when embedding rotated fonts inside a metafile.
	//curr_font->CreateFontIndirect(&m_font_table[font]);
	curr_font->CreatePointFontIndirect(&m_font_table[font]);

	// Do we need to record the actual font size?
	if(!(update_status & TEXT_SET_AVE_HEIGHT)){
		update_status |= TEXT_SET_AVE_HEIGHT;
		LOGFONT tmpFont;

		curr_font->GetLogFont(&tmpFont);
		
		ave_height = -MulDiv(tmpFont.lfHeight, 1, 3);
	}

	pDC->SetTextColor(color);

	// Save a pointer to the original font
	CFont *return_value = pDC->SelectObject(curr_font);

	// Delete the old font but remember its value
	if(old_ptr){
		// It is only safe to delete this pointer AFTER we have
		// selected curr_font. 
		delete old_ptr;
	}

	if(return_value == old_ptr){
		// We just replaced the current font with the SAME font!
		// This means that we just DELETED the return value -- so
		// return the freshly allocated font instead. 
		return curr_font;
	}

	return return_value;
}

void TextElem::Push(TextElem *pText, int loc, float font_scale /*= 1.0f*/)
{
	UpdateAll();

	if(pText == NULL)
		// Don't throw an error!
		//throw Mem_Error("Invalid pointer!", __FILE__, __LINE__);
		return;

	if((loc < TEXT_TOP) || (loc > TEXT_BOTTOM))
		throw Bounds_Error("loc out of bounds", __FILE__, __LINE__);

	if(next[loc])
		throw Error("Overwrite!", __FILE__, __LINE__);

	if(font_scale != 1.0f)
		pText->ScalePointSize(font_scale);

	next[loc] = pText;
}

TextElem* TextElem::GetNext(int m_loc)
{
	if((m_loc < TEXT_TOP ) || (m_loc > TEXT_BOTTOM))
		throw Bounds_Error("Index out of bounds!", __FILE__, __LINE__);

	return next[m_loc];
}

CRect TextElem::GetExtent(CDC *m_pDC, LOGFONT *m_font_table, 
						  int m_x /* = 0 */, int m_y /* = 0 */)
{
	CRect extent;

	int left_gap = MulDiv(LEFT_GAP, printScaleMul, printScaleDiv), 
		vertical_gap = MulDiv(VERTICAL_GAP, printScaleMul, printScaleDiv), 
		right_gap = MulDiv(RIGHT_GAP, printScaleMul, printScaleDiv);

	// The size of the top and bottom elements (if present)
	CSize topSize, bottomSize;
	int bottomOffset = 0, 
		midOffset =0, 
		topOffset = 0;

	// Is it safe to delete the current font pointer at the end of
	// this function? This will only hold true if curr_font has not
	// been previously allocated memory by another function. If we aren't
	// carefull to delete GDI objects (like CFont) that are no longer needed,
	// we could run out of system resources if too many TextElems are displayed
	// on the screen.
	BOOL safe_font_delete = (curr_font == NULL);

	// Select the font here ....
	CFont *old_font = SetFont(m_pDC, m_font_table);

	// Align all text along the baseline
	UINT old_alignment = m_pDC->SetTextAlign(PRESTO_TEXT_ALIGNMENT);

	int tmpOffset = 0;
	CSize tmp = GetTextExtent(m_pDC);
	
	// Compute the height
	tmp.cy = GetElemHeight(m_pDC, m_font_table, &tmpOffset);

	if(options & TEXT_ALIGN_MID){
		midOffset = -tmpOffset;
	}

	// If this element has top and/or bottom elements, let the largest width
	// set the extent.
	if(next[TEXT_TOP]){
		CRect topRect = next[TEXT_TOP]->GetExtent(m_pDC, m_font_table);

		// Save the size of the top element
		topSize.cx = topRect.Width();
		topSize.cy = topRect.Height();

		topOffset = topRect.bottom;

		tmp.cx = max(topSize.cx, tmp.cx);
	}

	if(next[TEXT_BOTTOM]){
		CRect bottomRect = next[TEXT_BOTTOM]->GetExtent(m_pDC, m_font_table);

		// Save the size of the bottom element
		bottomSize.cx = bottomRect.Width();
		bottomSize.cy = bottomRect.Height();

		bottomOffset = bottomRect.bottom;

		tmp.cx = max(bottomSize.cx, tmp.cx);
	}

	if(options & TEXT_FRAC_LINE){
		// Increment the height to make room for
		// the line we just added
		tmp.cy += MulDiv(FRAC_LINE_WIDTH, printScaleMul, printScaleDiv);
	}

	// Do we need to insert a midpoint offset?
	if(options & TEXT_ALIGN_MID){
		//tmpOffset += (tmp.cy - MulDiv(ave_height, printScaleMul, printScaleDiv))/2;
		tmpOffset += (tmp.cy - ave_height)/2;
	}

	// Construct the bounding rect
	extent.top = m_y - tmp.cy + tmpOffset + midOffset;
	extent.bottom = m_y + tmpOffset + midOffset;
	extent.left = m_x;
	extent.right = m_x + tmp.cx;

	// Try getting rid of this later
	extent.NormalizeRect();

	if(next[TEXT_TOP]){
		int x = (tmp.cx - topSize.cx)/2;
		int y = 0;

		if(options & TEXT_CENTER_THIS){
			y += - vertical_gap - tmp.cy - topOffset;
		}
		else{
			if(options & TEXT_CENTER_TOP){
				/*y = 0;*/
			}
			else{ // options & TEXT_CENTER_BOTTOM
				y += - vertical_gap - tmp.cy - bottomSize.cy;
			}
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		CRect tmpRect = 
			next[TEXT_TOP]->GetExtent(m_pDC, m_font_table, 
				m_x + x, m_y + y);

		extent.UnionRect(&extent, &tmpRect);
	}

	if(next[TEXT_SUPER_R]){
		int y = 0;

		if(tmp.cy == 0){
			if(options & TEXT_CENTER_THIS){
				//y += -(int)(m_pDC->GetTextExtent(" ").cy);
				y += -2*ave_height;
			}
			else{
				if(options & TEXT_CENTER_TOP){
					y += -topSize.cy;
				}
				else{ // options & TEXT_CENTER_BOTTOM
					y += -bottomSize.cy;
				}
			}
		}
		else
			y += -(int)(tmp.cy);

		if(options & TEXT_FRAC_LINE){
			y += -topSize.cy;
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		CRect tmpRect = 
			next[TEXT_SUPER_R]->GetExtent(m_pDC, m_font_table, 
				m_x + tmp.cx + right_gap, m_y + y);
		extent.UnionRect(&extent, &tmpRect);
	}

	if(next[TEXT_RIGHT]){
		int y = 0;

		// Allow an offset to make room for superscripts or
		// subscripts
		if(next[TEXT_SUPER_R] && !(options & TEXT_ALIGN_MID)){
			CRect tmpRect = next[TEXT_SUPER_R]->GetExtent(m_pDC, m_font_table);
			tmpOffset = tmpRect.Width();
		}
		else
			tmpOffset = 0;

		if(next[TEXT_SUB_R] && !(options & TEXT_ALIGN_MID)){
			CRect tmpRect = next[TEXT_SUB_R]->GetExtent(m_pDC, m_font_table);
			tmpOffset += tmpRect.Width();
		}

		CRect tmpRect = 
			next[TEXT_RIGHT]->GetExtent(m_pDC, m_font_table, 
				m_x + tmpOffset + tmp.cx + right_gap, m_y + y);
		extent.UnionRect(&extent, &tmpRect);
	}	

	if(next[TEXT_SUB_R]){
		int y = 0;

		if((tmp.cy == 0) && (options & TEXT_CENTER_BOTTOM)){
			y += bottomSize.cy;
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		CRect tmpRect = next[TEXT_SUB_R]->GetExtent(m_pDC, m_font_table, 
				m_x + tmp.cx + right_gap, m_y);

		y += tmpRect.Height()/2;

		if(options & TEXT_FRAC_LINE){
			y += bottomSize.cy;
		}

		tmpRect.bottom += y;
		tmpRect.top += y;
		
		extent.UnionRect(&extent, &tmpRect);
	}

	if(next[TEXT_BOTTOM]){
		int x = (tmp.cx - bottomSize.cx)/2;
		int y = 0;

		if(options & TEXT_CENTER_THIS){
			y += vertical_gap + bottomSize.cy - bottomOffset;
		}
		else{
			if(options & TEXT_CENTER_TOP){
				y += vertical_gap + tmp.cy + bottomSize.cy;
			}
			else{ // options & TEXT_CENTER_BOTTOM
				/*y = 0;*/
			}
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		CRect tmpRect = 
			next[TEXT_BOTTOM]->GetExtent(m_pDC, m_font_table, 
				m_x + x, m_y + y);
		extent.UnionRect(&extent, &tmpRect);
	}

	m_pDC->SetTextAlign(old_alignment);
	m_pDC->SelectObject(old_font);

	// Clean up unused GDI resources if they are no
	// longer needed.
	if(safe_font_delete && curr_font){
		delete curr_font;
		curr_font = NULL;
	}

	return extent;
}

CSize TextElem::GetTextExtent(CDC *m_pDC)
{
	if(update_status & TEXT_SET_EXTENT)
		return elem_extent;

	update_status |= TEXT_SET_EXTENT;
	elem_extent = m_pDC->GetTextExtent(elem);

	return elem_extent;
}

// Provide an accurate height and offset for the elem string with the 
// currently selected font.
int TextElem::GetElemHeight(CDC *m_pDC, LOGFONT *m_font_table, int *m_offset /* NULL */)
{
	// Only update the element height as needed!
	if(update_status & TEXT_SET_HEIGHT){
		if(m_offset){
			*m_offset = offset;
		}

		return elem_height;
	}

	update_status |= TEXT_SET_HEIGHT;

	int str_len = elem.GetLength();

	if(str_len == 0){
		// We have a zero length string, so don't bother
		// computing the text metrics.
		offset = 0;
		elem_height = 0;

		if(m_offset){
			*m_offset = offset;
		}

		return elem_height;
	}

	MAT2 rotation;
	int min_offset, max_offset;

	elem_height = 0;
	min_offset = -1;

	// The anti-rotation matrix (that allows us to compute
	// glyph demensions in an unrotated frame).
	rotation.eM11 = FloatToFixed(c);
	rotation.eM12 = FloatToFixed(-s);
	rotation.eM21 = FloatToFixed(s);
	rotation.eM22 = FloatToFixed(c);

	CFont *old_font = SetFont(m_pDC, m_font_table);

	for(int i = 0;i < str_len;i++){
		int ret_value;

		ret_value = m_pDC->GetGlyphOutline(
				elem[i],        // character to query
				0,				// format of data to return
				&gm,			// pointer to structure for metrics
				0,				// size of buffer for data
				NULL,			// pointer to buffer for data
				&rotation);		// pointer to transformation matrix structure
		
		if(ret_value != GDI_ERROR){
			elem_height = max(elem_height, (int)(gm.gmBlackBoxY));

			if(min_offset == -1){
				// Initialize offset values
				min_offset = max_offset = gm.gmBlackBoxY - gm.gmptGlyphOrigin.y;
			}
			else{
				// Use offset as a temporary variable here
				offset = gm.gmBlackBoxY - gm.gmptGlyphOrigin.y;
				
				// Compute the largest and smallest character offsets
				min_offset = min(min_offset, offset);
				max_offset = max(max_offset, offset);
			}
		}
		else{
			// An error has occured. Fall back to using GetTextMetrics.
			// This function only returns information about the average 
			// properties of the font -- not specific information about a given
			// glyph or character.
			elem_height = 0;

			TEXTMETRIC tm;

			if(m_pDC->GetTextMetrics(&tm)){
				elem_height = tm.tmHeight;
			}

			min_offset = max_offset = 0;
		}
	}

	// Restore the font
	m_pDC->SelectObject(old_font);

	if(min_offset == max_offset){
		offset = min_offset;
	}
	else{
		offset = max_offset;
		
		// If we have unequal offsets, we need to make more room for
		// the text string
		elem_height += max_offset - min_offset;
	}

	if(m_offset){
		*m_offset = offset;
	}

	return elem_height;
}

void TextElem::Draw(CDC *m_pDC, LOGFONT *m_font_table, int m_x, int m_y)
{
	// The size of the top and bottom elements (if present)
	CSize topSize, bottomSize;
	int bottomOffset = 0, 
		midOffset = 0, 
		topOffset = 0;

	int left_gap = MulDiv(LEFT_GAP, printScaleMul, printScaleDiv), 
		vertical_gap = MulDiv(VERTICAL_GAP, printScaleMul, printScaleDiv), 
		right_gap = MulDiv(RIGHT_GAP, printScaleMul, printScaleDiv);

	int x, y;
	
	// Is it safe to delete the current font pointer at the end of
	// this function? This will only hold true if curr_font has not
	// been previously allocated memory by another function. If we aren't
	// carefull to delete GDI objects (like CFont) that are no longer needed,
	// we could run out of system resources if too many TextElems are displayed
	// on the screen.
	BOOL safe_font_delete = (curr_font == NULL);

	// Select the font here ....
	CFont *old_font = SetFont(m_pDC, m_font_table);

	// Align all text along the baseline
	UINT old_alignment = m_pDC->SetTextAlign(PRESTO_TEXT_ALIGNMENT);

	// Set the background drawing mode
	int old_bk_mode = m_pDC->SetBkMode(TRANSPARENT);

	CSize tmp = GetTextExtent(m_pDC);

	if(options & TEXT_ALIGN_MID){
		tmp.cy = GetElemHeight(m_pDC, m_font_table, &midOffset);
		midOffset = -midOffset;
	}
	else
		tmp.cy = GetElemHeight(m_pDC, m_font_table);

	// If this element has top and/or bottom elements, let the largest width
	// set the extent.
	if(next[TEXT_TOP]){
		CRect topRect = next[TEXT_TOP]->GetExtent(m_pDC, m_font_table);

		// Save the size of the top element
		topSize.cx = topRect.Width();
		topSize.cy = topRect.Height();

		topOffset = topRect.bottom;

		tmp.cx = max(topSize.cx, tmp.cx);
	}

	if(next[TEXT_BOTTOM]){
		CRect bottomRect = next[TEXT_BOTTOM]->GetExtent(m_pDC, m_font_table);

		// Save the size of the bottom element
		bottomSize.cx = bottomRect.Width();
		bottomSize.cy = bottomRect.Height();

		bottomOffset = bottomRect.bottom;

		tmp.cx = max(bottomSize.cx, tmp.cx);
	}

	if(options & TEXT_FRAC_LINE){
		int line_width = MulDiv(FRAC_LINE_WIDTH, printScaleMul, printScaleDiv);

		int tmp_x = tmp.cx,
			tmp_y = -ave_height/2 - FRAC_LINE_WIDTH;

		// We need to define a brush in order to use the advanced CPen
		// constructor (that permits the specification of endcape).
		LOGBRUSH frac_brush;
		frac_brush.lbColor = RGB(0, 0, 0);
		frac_brush.lbStyle = BS_SOLID;

		CPen frac_pen(PS_GEOMETRIC | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL, 
			line_width, 
			&frac_brush,
			0,
			NULL);

		CPen *old_pen;

		old_pen = m_pDC->SelectObject(&frac_pen);
		
		// Draw the dividing line
		m_pDC->MoveTo(m_x + (int)(s*tmp_y), m_y + (int)(c*tmp_y));
		m_pDC->LineTo(m_x + (int)(c*tmp_x + s*tmp_y), m_y + (int)(-s*tmp_x + c*tmp_y));

		// Restore the original pen
		m_pDC->SelectObject(old_pen);

		// Increment the height to make room for
		// the line we just added
		tmp.cy += line_width;
	}

	// Draw the text
	if(options & TEXT_ALIGN_MID){
		//midOffset += (tmp.cy - MulDiv(ave_height, printScaleMul, printScaleDiv))/2;
		midOffset += (tmp.cy - ave_height)/2;

		m_pDC->TextOut(m_x + (int)(s*midOffset), m_y + (int)(c*midOffset), elem);
	}
	else{
		m_pDC->TextOut(m_x, m_y, elem);
	}

	m_pDC->SetBkMode(old_bk_mode);

	if(next[TEXT_TOP]){
		x = (tmp.cx - topSize.cx)/2;
		y = 0;

		if(options & TEXT_CENTER_THIS){
			y += - vertical_gap - tmp.cy - topOffset;
		}
		else{
			if(options & TEXT_CENTER_TOP){
				/*y = 0;*/
			}
			else{ // options & TEXT_CENTER_BOTTOM
				y += - vertical_gap - tmp.cy - bottomSize.cy - topOffset;
			}
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		next[TEXT_TOP]->Draw(m_pDC, m_font_table, 
			m_x + (int)(c*x + s*y), m_y + (int)(-s*x + c*y));
	}

	if(next[TEXT_SUPER_R]){
		x = tmp.cx;
		y = 0;

		if(tmp.cy == 0){
			if(options & TEXT_CENTER_THIS){
				//y += -(int)(m_pDC->GetTextExtent(" ").cy);
				y += -2*ave_height;
			}
			else{
				if(options & TEXT_CENTER_TOP){
					y += -topSize.cy;
				}
				else{ // options & TEXT_CENTER_BOTTOM
					y += -bottomSize.cy;
				}
			}
		}
		else
			y += -tmp.cy;

		if(options & TEXT_FRAC_LINE){
			y += -topSize.cy;
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		next[TEXT_SUPER_R]->Draw(m_pDC, m_font_table, 
			m_x + (int)(c*x + s*y), m_y + (int)(-s*x + c*y));
	}

	if(next[TEXT_RIGHT]){
		x = tmp.cx + right_gap;
		y = 0;

		// Allow an offset to make room for superscripts or
		// subscripts
		if(next[TEXT_SUPER_R] && !(options & TEXT_ALIGN_MID)){
			CRect tmpRect = next[TEXT_SUPER_R]->GetExtent(m_pDC, m_font_table);
			x += tmpRect.Width();
		}

		if(next[TEXT_SUB_R] && !(options & TEXT_ALIGN_MID)){
			CRect tmpRect = next[TEXT_SUB_R]->GetExtent(m_pDC, m_font_table);
			x += tmpRect.Width();
		}

		next[TEXT_RIGHT]->Draw(m_pDC, m_font_table, 
			m_x + (int)(c*x + s*y), m_y + (int)(-s*x + c*y));
	}	

	if(next[TEXT_SUB_R]){
		x = tmp.cx;
		y = 0;

		if((tmp.cy == 0) && (options & TEXT_CENTER_BOTTOM)){
			y += bottomSize.cy;
		}

		CRect tmpRect = next[TEXT_SUB_R]->GetExtent(m_pDC, m_font_table, 
				m_x + tmp.cx + right_gap, m_y);

		y += tmpRect.Height()/2;

		if(options & TEXT_FRAC_LINE){
			y += bottomSize.cy;
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		next[TEXT_SUB_R]->Draw(m_pDC, m_font_table, 
			m_x + (int)(c*x + s*y), m_y + (int)(-s*x + c*y));
	}

	if(next[TEXT_BOTTOM]){
		x = (tmp.cx - bottomSize.cx)/2;
		y = 0;

		if(options & TEXT_CENTER_THIS){
			y += vertical_gap + bottomSize.cy - bottomOffset;
		}
		else{
			if(options & TEXT_CENTER_TOP){
				y += vertical_gap + tmp.cy + bottomSize.cy;
			}
			// options & TEXT_CENTER_BOTTOM
			//else{ 
			//	y = 0;
			//}
		}

		// Do we need to insert a midpoint offset?
		if(options & TEXT_ALIGN_MID){
			y += (tmp.cy - ave_height)/2;
		}

		next[TEXT_BOTTOM]->Draw(m_pDC, m_font_table, 
			m_x + (int)(c*x + s*y), m_y + (int)(-s*x + c*y));
	}

	// Align all text along the baseline
	m_pDC->SetTextAlign(old_alignment);

	// Restore the original font
	m_pDC->SelectObject(old_font);

	if(safe_font_delete && curr_font){
		delete curr_font;
		curr_font = NULL;
	}
}

TextLabel::TextLabel()
{
	label = NULL;

	num_font = PREST_FONT_COUNT;
	font_table = new LOGFONT[num_font];

	// Set the font parameters
	for(int i = 0;i < num_font;i++){
		font_table[i].lfHeight = DefaultTextFontSize;
		font_table[i].lfWidth = 0;
		font_table[i].lfEscapement = 0;
		font_table[i].lfOrientation = font_table[i].lfEscapement;
		font_table[i].lfWeight = FW_NORMAL;
		font_table[i].lfItalic = FALSE;
		font_table[i].lfUnderline = FALSE;
		font_table[i].lfStrikeOut = FALSE;

		// Allow the the lfFaceName parameter to specify the 
		// charset to use.
		font_table[i].lfCharSet = DEFAULT_CHARSET;
		font_table[i].lfOutPrecision = OUT_DEFAULT_PRECIS;
		font_table[i].lfClipPrecision = CLIP_DEFAULT_PRECIS;
		font_table[i].lfQuality = DEFAULT_QUALITY;
		font_table[i].lfPitchAndFamily = FF_DONTCARE|DEFAULT_PITCH;
	}

	// Mimic LaTeX by making the math font italic and always in a western
	// script.
	font_table[PRESTO_FONT_MATH].lfItalic = TRUE;
	font_table[PRESTO_FONT_MATH].lfCharSet = ANSI_CHARSET;

	// The face name listed here should match DefaultFont defined in sgraph.h
	// (for a consistent "look and feel").
	lstrcpy(font_table[PRESTO_FONT_DEFAULT].lfFaceName, "Times New Roman");
	lstrcpy(font_table[PRESTO_FONT_MATH].lfFaceName, "Times New Roman");

	lstrcpy(font_table[PRESTO_FONT_SYMBOL].lfFaceName, "Symbol");

	// Not every computer may have this font!
	lstrcpy(font_table[PRESTO_FONT_SYMBOL_EXT1].lfFaceName, "Math3");
}

TextLabel::~TextLabel()
{
	if(label){
		delete label;
		label = NULL;
	}

	if(font_table){
		delete [] font_table;
		font_table = NULL;
	}
}

void TextLabel::SetLabel(CString &m_text, LOGFONT &m_font, COLORREF m_color)
{
	if(label){
		delete label;
		label = NULL;
	}

	color = m_color;
	angle = m_font.lfEscapement;

	for(int i = 0;i < num_font;i++){
		font_table[i].lfHeight = m_font.lfHeight;
		font_table[i].lfEscapement = angle;
		font_table[i].lfOrientation = angle;
	}

	lstrcpy(font_table[PRESTO_FONT_DEFAULT].lfFaceName, 
		m_font.lfFaceName);

	lstrcpy(font_table[PRESTO_FONT_MATH].lfFaceName, 
		m_font.lfFaceName);

	pt_size = -m_font.lfHeight;

	//////////////////////////// Parse input string //////////////////////////// 
	if(!Parse(m_text)){
		// If we can't parse the test, use the input string verbatim
		if(label)
			delete label;

		label = new TextElem(m_text, PRESTO_FONT_DEFAULT, pt_size, 
			angle, color);
	}
}

void TextLabel::Draw(CDC *m_pDC, int m_x, int m_y)
{
	if(label){
		CRect extent = label->GetExtent(m_pDC, font_table);

		// Map xy into the rotated, shifted coordinate frame:
		//     |  c  s | | x + off_x| =  | x_p | 
		//     | -s  c | | y + off_y|    | y_p |
		float c = (float)(cos(angle*1.74532925199e-3f));
		float s = (float)(sin(angle*1.74532925199e-3f));

		int dx = (extent.right + extent.left)/2;
		int dy = (extent.top + extent.bottom)/2;

		int x_p;
		int y_p;

		x_p = m_x - (int)(c*dx + s*dy);
		y_p = m_y - (int)(-s*dx + c*dy);
		
		label->Draw(m_pDC, font_table, x_p, y_p);
	}
}

CRect TextLabel::GetExtent(CDC *m_pDC)
{
	if(!label)
		return CRect(0, 0, 0, 0);
	
	return label->GetExtent(m_pDC, font_table);
}


void TextLabel::Copy(const TextLabel& m_tl)
{
	if(label){
		delete label;
		label = NULL;
	}
	
	if(m_tl.label){
		label = new TextElem;
		label->Copy(*m_tl.label);
	}

	if(font_table){
		delete [] font_table;
		font_table = NULL;
	}
	
	angle = m_tl.angle;

	num_font = m_tl.num_font;

	font_table = new LOGFONT[num_font];

	for(int i = 0;i < num_font;i++)
		font_table[i] = m_tl.font_table[i];
}

// Return TRUE if the input string has been successfully parsed,
// FALSE otherwise
BOOL TextLabel::Parse(CString &m_text)
{
	if(m_text.GetLength() == 0)
		// Empty string, nothing to do!
		return TRUE;

	// Make a copy of the input string
	char *buffer = new char [m_text.GetLength() + 1];
	
	if(!buffer)
		throw Mem_Error("Can't allocate text buffer", __FILE__, __LINE__);

	// Copy the buffer and omit any '\r' or '\n' characters for now ...
	//strcpy(buffer, LPCSTR(m_text));
	int copy_count = 0;
	int i;
	for(i = 0;i < m_text.GetLength();i++)
		if((m_text[i] != '\r') && (m_text[i] != '\n'))
			buffer[copy_count++] = m_text[i];

	// Terminate the string
	buffer[copy_count] = '\0';

	list<TextToken> token_list;

	// Break the input string into tokens
	if(!GetTokenList(buffer, token_list)){
		delete [] buffer;

		return FALSE;
	}

	TextElem *curElem;

	for(i = 0;i < token_list.Length();i++){
		TextElem *nextElem;
		TextToken nextToken = token_list[i];

		if(nextToken.latex){
			if((nextElem = StringToLatex(nextToken.token)) == NULL)
				// If the string can not be parsed as a LaTex string,
				// diplay it as plain text.
				nextElem = new TextElem(CString(nextToken.token), 
					PRESTO_FONT_DEFAULT, pt_size, angle, color);
		}
		else{
			// Handle plain old text
			nextElem = new TextElem(CString(nextToken.token), 
				PRESTO_FONT_DEFAULT, pt_size, angle, color);
		}

		if(nextElem == NULL)
			throw Mem_Error("Can't allocate text elem", __FILE__, __LINE__);

		if(label == NULL){
			label = nextElem;
			curElem = label;
		}
		else{
			// Append text in the right-most position
			while(curElem->GetNext(TEXT_RIGHT))
				curElem = curElem->GetNext(TEXT_RIGHT);

			curElem->Push(nextElem, TEXT_RIGHT);
			curElem = curElem->GetNext(TEXT_RIGHT);
		}
	}
	
	// Clean up
	delete [] buffer;

	return TRUE;
}

TextElem* TextLabel::StringToLatex(char *buffer)
{
	static char mask[] = "\t \\{}()^_,[]<>.;:+-=*/|";
	static char whitespace[] = "\t \n";

	// A list of characters that can be proceeded by
	// an escape (i.e. '\') character
	static char escape_mask[] = "{}_$&#%";

	if(buffer == NULL)
		throw Error("Bad buffer", __FILE__, __LINE__);

	// Throw away any white space
	buffer = buffer + strspn(buffer, whitespace);

	if(strlen(buffer) == 0){
		return new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
			angle, color);
	}

	// A default element
	TextElem *newElem = NULL;
	TextElem *curElem = NULL;
	TextElem *tmpElem = NULL;

	char *tmp;
	int len;
	int new_loc = -1;

	/////////////////////////// Test for base-line text ///////////////////////////
	switch(buffer[0]){
		case TEXT_ESCAPE:
			buffer = buffer + 1;

			if(strchr(escape_mask, buffer[0]))
				// This is a single escaped character
				len = 1;
			else
				// Read the escape sequence
				len = strcspn(buffer, mask);

			newElem = new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
				angle, color);

			len = GetEscape(newElem, buffer, len);

			if(len < 0){
				delete newElem;
				return NULL;
			}

			buffer = buffer + len;
			break;
		case '{':
			// Find the matching brace (not a sophisticated test -- this will
			// need improvement).
			//tmp = strchr(buffer + 1, '}');
			tmp = MatchCurlyBrace(buffer + 1);

			// If there's no matching brace, it's an error
			if(tmp == NULL)
				return NULL;
			
			tmp[0] = '\0';

			newElem = StringToLatex(buffer + 1);

			if(newElem == NULL)
				return NULL;

			buffer = tmp + 1;
			break;
		case '}':
			return NULL;
			break;
		case '_':
			return NULL;
			break;
		case '^':
			return NULL;
			break;
		default:
			newElem = new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
				angle, color);

			// Write characters in math font ... 
			newElem->SetString(buffer, 1);

			// ... unless the character is a '(', ')', '-', '/', '+', etc.
			if(!isalpha(buffer[0]))
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);

			buffer = buffer + 1;
			break;
	};

	/////////////////////////// Test for super or subscript text ///////////////////////////

	if(newElem == NULL)
		throw Error("Latex Parse Error", __FILE__, __LINE__);

	// Throw away any white space
	buffer = buffer + strspn(buffer, whitespace);

	// If there is no more text to parse, return
	if(strlen(buffer) == 0)
		return newElem;

	// Append any further text in the right-most position
	curElem = newElem;
	while(curElem->GetNext(TEXT_RIGHT))
		curElem = curElem->GetNext(TEXT_RIGHT);

	// Is the next chacter a subscript or superscript?
	if((buffer[0] == '_') || (buffer[0] == '^')){
		if(buffer[0] == '_'){
			if(curElem->GetOptions() & TEXT_SUB_BOTTOM)
				new_loc = TEXT_BOTTOM;
			else
				new_loc = TEXT_SUB_R;
		}
		else{
			if(curElem->GetOptions() & TEXT_SUPER_TOP)
				new_loc = TEXT_TOP;
			else
				new_loc = TEXT_SUPER_R;
		}

		// Remove the '_' or '^' character
		buffer = buffer + 1;

		switch(buffer[0]){
			case '{':
				// Find the matching brace (not a sophisticated test -- this will
				// need improvement).
				//tmp = strchr(buffer + 1, '}');
				tmp = MatchCurlyBrace(buffer + 1);

				// If there's no matching brace, it's an error
				if(tmp == NULL){
					delete newElem;
					return NULL;
				}

				tmp[0] = '\0';
	
				curElem->Push(StringToLatex(buffer + 1), new_loc, SUPER_SUB_SCALE);
				buffer = tmp + 1;
				break;
			case '\\':
				
				buffer = buffer + 1;

				if(strchr(escape_mask, buffer[0]))
					// This is a single escaped character
					len = 1;
				else
					// Read the escape sequence
					len = strcspn(buffer, mask);

				tmpElem = new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
					angle, color);
				
				len = GetEscape(tmpElem, buffer, len);

				if(len < 0){
					delete tmpElem;
					delete newElem;
					return NULL;
				}

				buffer = buffer + len;

				curElem->Push(tmpElem, new_loc, SUPER_SUB_SCALE);
				break;
			default:
				tmpElem = new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
					angle, color);

				// Write characters in math font ...
				tmpElem->SetString(buffer, 1);
				
				// ... unless the character is a '(', ')', '-', '/', '+', etc.
				if(!isalpha(buffer[0]))
					tmpElem->SetFontIndex(PRESTO_FONT_DEFAULT);

				buffer = buffer + 1;

				curElem->Push(tmpElem, new_loc, SUPER_SUB_SCALE);
				break;
		};
	}

	/////////////////////////// Test for super or subscript text ///////////////////////////

	if(newElem == NULL)
		throw Error("Latex Parse Error", __FILE__, __LINE__);

	// Throw away any white space
	buffer = buffer + strspn(buffer, whitespace);

	// If there is no more text to parse, return
	if(strlen(buffer) == 0)
		return newElem;

	// Append any further text in the right-most position
	curElem = newElem;
	while(curElem->GetNext(TEXT_RIGHT))
		curElem = curElem->GetNext(TEXT_RIGHT);

	// Is the next chacter a subscript or superscript?
	if((buffer[0] == '_') || (buffer[0] == '^')){
		if(buffer[0] == '_'){
			// Have we already read a subscript
			if(new_loc == TEXT_SUB_R){
				delete newElem;
				return NULL;
			}

			if(curElem->GetOptions() & TEXT_SUB_BOTTOM)
				new_loc = TEXT_BOTTOM;
			else
				new_loc = TEXT_SUB_R;
		}
		else{
			// Have we already read a superscript
			if(new_loc == TEXT_SUPER_R){
				delete newElem;
				return NULL;
			}

			if(curElem->GetOptions() & TEXT_SUPER_TOP)
				new_loc = TEXT_TOP;
			else
				new_loc = TEXT_SUPER_R;
		}

		// Remove the '_' or '^' character
		buffer = buffer + 1;

		switch(buffer[0]){
			case '{':
				// Find the matching brace (not a sophisticated test -- this will
				// need improvement).
				//tmp = strchr(buffer + 1, '}');
				tmp = MatchCurlyBrace(buffer + 1);

				// If there's no matching brace, it's an error
				if(tmp == NULL){
					delete newElem;
					return NULL;
				}

				tmp[0] = '\0';
	
				curElem->Push(StringToLatex(buffer + 1), new_loc, SUPER_SUB_SCALE);
				buffer = tmp + 1;
				break;
			case '\\':
				buffer = buffer + 1;

				if(strchr(escape_mask, buffer[0]))
					// This is a single escaped character
					len = 1;
				else
					// Read the escape sequence
					len = strcspn(buffer, mask);

				tmpElem = new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
					angle, color);

				len = GetEscape(tmpElem, buffer, len);

				if(len < 0){
					delete tmpElem;
					delete newElem;
					return NULL;
				}

				buffer = buffer + len;

				curElem->Push(tmpElem, new_loc, SUPER_SUB_SCALE);
				break;
			default:
				tmpElem = new TextElem(CString(""), PRESTO_FONT_MATH, pt_size, 
					angle, color);

				// Write characters in math font ...
				tmpElem->SetString(buffer, 1);

				// ... unless the character is a '(', ')', '-', '/', '+', etc.
				if(!isalpha(buffer[0]))
					tmpElem->SetFontIndex(PRESTO_FONT_DEFAULT);

				buffer = buffer + 1;

				curElem->Push(tmpElem, new_loc, SUPER_SUB_SCALE);
				break;
		};
	}

	// Add any remaining text to the right
	curElem = newElem;
	while(curElem->GetNext(TEXT_RIGHT))
		curElem = curElem->GetNext(TEXT_RIGHT);

	curElem->Push(StringToLatex(buffer), TEXT_RIGHT);

	return newElem;
}

// Return -1 on failure, otherwise return the length of the string
int TextLabel::GetEscape(TextElem *newElem, char *buffer, int len)
{
	if(buffer == NULL)
		throw Mem_Error("Null buffer!", __FILE__, __LINE__);

	if(len == 0)
		return -1;

	if(isupper(buffer[0]))
		switch(buffer[0]){
			// The upper case greek letters
			case 'G': // Gamma
				if((len == 5) && (strncmp("Gamma", buffer, len) == 0)){
					newElem->SetString("G");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'D': // Delta, Downarrow
				if((len == 5) && (strncmp("Delta", buffer, len) == 0)){
					newElem->SetString("D");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 9) && (strncmp("Downarrow", buffer, len) == 0)){
					newElem->SetString((char)0xDF);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 'T': // Theta
				if((len == 5) && (strncmp("Theta", buffer, len) == 0)){
					newElem->SetString("T");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'L': // Lambda, Leftarrow, Leftrightarrow
				if((len == 6) && (strncmp("Lambda", buffer, len) == 0)){
					newElem->SetString("L");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 9) && (strncmp("Leftarrow", buffer, len) == 0)){
					newElem->SetString((char)0xDC);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 14) && (strncmp("Leftrightarrow", buffer, len) == 0)){
					newElem->SetString((char)0xDB);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 'X': // Xi
				if((len == 2) && (strncmp("Xi", buffer, len) == 0)){
					newElem->SetString("X");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'P': // Pi, Phi, Psi, Pr
				if((len == 2) && (strncmp("Pi", buffer, len) == 0)){
					newElem->SetString("P");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("Phi", buffer, len) == 0)){
					newElem->SetString("F");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("Psi", buffer, len) == 0)){
					newElem->SetString("Y");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 2) && (strncmp("Pr", buffer, len) == 0)){
					newElem->SetString("Pr");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				return -1;
				break;
			case 'S': // Sigma
				if((len == 5) && (strncmp("Sigma", buffer, len) == 0)){
					newElem->SetString("S");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'U': // Upsilon
				if((len == 7) && (strncmp("Upsilon", buffer, len) == 0)){
					newElem->SetString("U");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'O': // Omega
				if((len == 5) && (strncmp("Omega", buffer, len) == 0)){
					newElem->SetString("W");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'R': // Rightarrow
				if((len == 10) && (strncmp("Rightarrow", buffer, len) == 0)){
					newElem->SetString((char)0xDE);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'A': // AA
				if((len == 2) && (strncmp("AA", buffer, len) == 0)){
					newElem->SetString((char)0xDE);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL_EXT1);
					return len;
				}
				else
					return -1;
				break;
		};
	
	if(islower(buffer[0]))
		switch(buffer[0]){
		// The upper case greek letters
			case 'a': // alpha, approx, arccos, arcsin, arctan, arg
				if((len == 5) && (strncmp("alpha", buffer, len) == 0)){
					newElem->SetString("a");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 6) && (strncmp("approx", buffer, len) == 0)){
					newElem->SetString((char)0xBB);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("arccos", buffer, len) == 0)){
					newElem->SetString("arccos");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("arcsin", buffer, len) == 0)){
					newElem->SetString("arcsin");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("arctan", buffer, len) == 0)){
					newElem->SetString("arctan");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("arg", buffer, len) == 0)){
					newElem->SetString("arg");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}
				return -1;
				break;
			case 'b': // beta, bullet, bar{}
				if((len == 4) && (strncmp("beta", buffer, len) == 0)){
					newElem->SetString("b");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 6) && (strncmp("bullet", buffer, len) == 0)){
					newElem->SetString((char)0xB7);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("bar", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = StackString(newElem, buffer + len, (char)(0x2D) /* - */, 
						PRESTO_FONT_SYMBOL, 0.5f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				break;
			case 'g': // gamma, geq, gcd
				if((len == 5) && (strncmp("gamma", buffer, len) == 0)){
					newElem->SetString("g");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("geq", buffer, len) == 0)){
					newElem->SetString((char)0xB3);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("gcd", buffer, len) == 0)){
					newElem->SetString("gcd");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				return -1;
				break;
			case 'h': // hom, hbar, hat{}
				if((len == 3) && (strncmp("hom", buffer, len) == 0)){
					newElem->SetString("hom");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 4) && (strncmp("hbar", buffer, len) == 0)){
					newElem->SetString((char)0xD1);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL_EXT1);
					return len;
				}

				if((len == 3) && (strncmp("hat", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = StackString(newElem, buffer + len, (char)(0xD9) /* ^ */, 
						PRESTO_FONT_SYMBOL, 0.5f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				return -1;
				break;
			case 'd': // delta, div, downarrow, deg, det, dim, dot{}, ddot{}
				if((len == 5) && (strncmp("delta", buffer, len) == 0)){
					newElem->SetString("d");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("div", buffer, len) == 0)){
					newElem->SetString((char)0xB8);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 9) && (strncmp("downarrow", buffer, len) == 0)){
					newElem->SetString((char)0xAF);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("deg", buffer, len) == 0)){
					newElem->SetString("deg");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("det", buffer, len) == 0)){
					newElem->SetString("det");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("dim", buffer, len) == 0)){
					newElem->SetString("dim");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("dot", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = StackString(newElem, buffer + len, (char)(0xB7) /* . */, 
						PRESTO_FONT_SYMBOL, 0.5f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				if((len == 4) && (strncmp("ddot", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					char tmpString[3];
					tmpString[0] = (char)(0xB7);
					tmpString[1] = (char)(0xB7);
					tmpString[2] = '\0';

					tmp = StackString(newElem, buffer + len, tmpString /* .. */, 
						PRESTO_FONT_SYMBOL, 0.5f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				return -1;
				break;
			case 'e': // epsilon, eta, equiv, exists, exp
				if((len == 7) && (strncmp("epsilon", buffer, len) == 0)){
					newElem->SetString("e");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("eta", buffer, len) == 0)){
					newElem->SetString("h");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 5) && (strncmp("equiv", buffer, len) == 0)){
					newElem->SetString((char)0xBA);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 6) && (strncmp("exists", buffer, len) == 0)){
					newElem->SetString((char)0x24);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("exp", buffer, len) == 0)){
					newElem->SetString("exp");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}
				return -1;
				break;
			case 'v': // varepsilon, vartheta, varpi, varrho, varsigma, varphi, vec{}
				if((len == 3) && (strncmp("vec", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = StackString(newElem, buffer + len, (char)(0xAE) /* -> */, 
						PRESTO_FONT_SYMBOL, 0.5f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				// Note that the Symbol TrueType font does not have a 
				// varepsilon character! Substitute epsilon in its place.
				if((len == 10) && (strncmp("varepsilon", buffer, len) == 0)){
					newElem->SetString("e");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				// Note that the Symbol TrueType font does not have a 
				// vartheta character! Substitute theta in its place.
				if((len == 8) && (strncmp("vartheta", buffer, len) == 0)){
					newElem->SetString("t");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				// Note that the Symbol TrueType font does not have a 
				// varpi character! Substitute pi in its place.
				if((len == 5) && (strncmp("varpi", buffer, len) == 0)){
					newElem->SetString("p");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				// Note that the Symbol TrueType font does not have a 
				// varrho character! Substitute rho in its place.
				if((len == 6) && (strncmp("varrho", buffer, len) == 0)){
					newElem->SetString("r");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				// Note that the Symbol TrueType font does not have a 
				// varsigma character! Substitute sigma in its place.
				if((len == 8) && (strncmp("varsigma", buffer, len) == 0)){
					newElem->SetString("s");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				// Note that the Symbol TrueType font does not have a 
				// varphi character! Substitute phi in its place.
				if((len == 6) && (strncmp("varphi", buffer, len) == 0)){
					newElem->SetString("f");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				
				break;
			case 'z': // zeta
				if((len == 5) && (strncmp("zeta", buffer, len) == 0)){
					newElem->SetString("z");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 't': // theta, tau, times, tan, tanh, tilde{}
				if((len == 5) && (strncmp("theta", buffer, len) == 0)){
					newElem->SetString("q");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("tau", buffer, len) == 0)){
					newElem->SetString("t");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 5) && (strncmp("times", buffer, len) == 0)){
					newElem->SetString((char)0xB4);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("tan", buffer, len) == 0)){
					newElem->SetString("tan");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 4) && (strncmp("tanh", buffer, len) == 0)){
					newElem->SetString("tanh");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 5) && (strncmp("tilde", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = StackString(newElem, buffer + len, (char)(0x7E) /* ~ */, 
						PRESTO_FONT_SYMBOL, 1.0f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				return -1;

				break;
			case 'i': // iota, infty, inf, int
				if((len == 4) && (strncmp("iota", buffer, len) == 0)){
					newElem->SetString("i");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 5) && (strncmp("infty", buffer, len) == 0)){
					newElem->SetString((char)0xA5);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("inf", buffer, len) == 0)){
					newElem->SetString("inf");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("int", buffer, len) == 0)){
					newElem->SetString((char)0xF2);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					newElem->ScalePointSize(2.0f);
					newElem->AddOptions(TEXT_ALIGN_MID);
					return len;
				}

				return -1;
				break;
			case 'k': // kappa, ker
				if((len == 5) && (strncmp("kappa", buffer, len) == 0)){
					newElem->SetString("k");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("ker", buffer, len) == 0)){
					newElem->SetString("ker");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				return -1;
				break;
			case 'l': // lambda, leq, leftarrow, leftrightarrow, lg, lim, 
					  // liminf, limsup, ln, log, ldots
				if((len == 6) && (strncmp("lambda", buffer, len) == 0)){
					newElem->SetString("l");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("leq", buffer, len) == 0)){
					newElem->SetString((char)0xA3);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 9) && (strncmp("leftarrow", buffer, len) == 0)){
					newElem->SetString((char)0xAC);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 14) && (strncmp("leftrightarrow", buffer, len) == 0)){
					newElem->SetString((char)0xAB);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 2) && (strncmp("lg", buffer, len) == 0)){
					newElem->SetString("lg");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("lim", buffer, len) == 0)){
					newElem->SetString("lim");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					newElem->AddOptions(TEXT_SUB_BOTTOM);
					return len;
				}
				 
				if((len == 6) && (strncmp("liminf", buffer, len) == 0)){
					newElem->SetString("liminf");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 6) && (strncmp("limsup", buffer, len) == 0)){
					newElem->SetString("limsup");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 2) && (strncmp("ln", buffer, len) == 0)){
					newElem->SetString("ln");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("log", buffer, len) == 0)){
					newElem->SetString("log");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 5) && (strncmp("ldots", buffer, len) == 0)){
					newElem->SetString((char)0xBC);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 'm': // mu, max, min
				if((len == 2) && (strncmp("mu", buffer, len) == 0)){
					newElem->SetString("m");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("max", buffer, len) == 0)){
					newElem->SetString("max");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("min", buffer, len) == 0)){
					newElem->SetString("min");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}
				return -1;
				break;
			case 'n': // nu, neq, nabla
				if((len == 2) && (strncmp("nu", buffer, len) == 0)){
					newElem->SetString("n");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("neq", buffer, len) == 0)){
					newElem->SetString((char)0xB9);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 5) && (strncmp("nabla", buffer, len) == 0)){
					newElem->SetString((char)0xD1);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 'x': // xi
				if((len == 2) && (strncmp("xi", buffer, len) == 0)){
					newElem->SetString("x");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				else
					return -1;
				break;
			case 'p': // pi, phi, psi, pm, propto, partial, perp, parallel
				if((len == 2) && (strncmp("pi", buffer, len) == 0)){
					newElem->SetString("p");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("phi", buffer, len) == 0)){
					newElem->SetString("f");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("psi", buffer, len) == 0)){
					newElem->SetString("y");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 2) && (strncmp("pm", buffer, len) == 0)){
					newElem->SetString((char)0xB1);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 6) && (strncmp("propto", buffer, len) == 0)){
					newElem->SetString((char)0xB5);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 7) && (strncmp("partial", buffer, len) == 0)){
					newElem->SetString((char)0xB6);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 4) && (strncmp("perp", buffer, len) == 0)){
					newElem->SetString((char)0x5E);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 8) && (strncmp("parallel", buffer, len) == 0)){
					newElem->SetString((char)0x37);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL_EXT1);
					return len;
				}
				return -1;
				break;
			case 'r': // rho, rightarrow
				if((len == 3) && (strncmp("rho", buffer, len) == 0)){
					newElem->SetString("r");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 10) && (strncmp("rightarrow", buffer, len) == 0)){
					newElem->SetString((char)0xAE);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 's': // sigma, sim, sec, sin, sinh, sup, sum, stackrel{}{}
				if((len == 5) && (strncmp("sigma", buffer, len) == 0)){
					newElem->SetString("s");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 3) && (strncmp("sum", buffer, len) == 0)){
					newElem->SetString("S");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					newElem->ScalePointSize(2.0f);
					newElem->AddOptions(TEXT_ALIGN_MID | TEXT_SUPER_TOP | 
										TEXT_SUB_BOTTOM);
					return len;
				}

				if((len == 3) && (strncmp("sim", buffer, len) == 0)){
					newElem->SetString((char)0x7E);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("sec", buffer, len) == 0)){
					newElem->SetString("sec");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("sin", buffer, len) == 0)){
					newElem->SetString("sin");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 4) && (strncmp("sinh", buffer, len) == 0)){
					newElem->SetString("sinh");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("sup", buffer, len) == 0)){
					newElem->SetString("sup");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 8) && (strncmp("stackrel", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = StackString2(newElem, buffer + len, 0.75f);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				return -1;
				break;
			case 'u': // upsilon, uparrow
				if((len == 7) && (strncmp("upsilon", buffer, len) == 0)){
					newElem->SetString("u");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 7) && (strncmp("uparrow", buffer, len) == 0)){
					newElem->SetString((char)0xAD);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 'c': // chi, cdot, cap, cup, cong, cos, cosh, cot, coth, csc
				if((len == 3) && (strncmp("chi", buffer, len) == 0)){
					newElem->SetString("c");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 4) && (strncmp("cdot", buffer, len) == 0)){
					newElem->SetString((char)0xD7);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("cap", buffer, len) == 0)){
					newElem->SetString((char)0xC7);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("cup", buffer, len) == 0)){
					newElem->SetString((char)0xC8);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 4) && (strncmp("cong", buffer, len) == 0)){
					newElem->SetString((char)0x40);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 3) && (strncmp("cos", buffer, len) == 0)){
					newElem->SetString("cos");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 4) && (strncmp("cosh", buffer, len) == 0)){
					newElem->SetString("cosh");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("cot", buffer, len) == 0)){
					newElem->SetString("cot");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 3) && (strncmp("csc", buffer, len) == 0)){
					newElem->SetString("csc");
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}
				return -1;
				break;
			case 'o': // omega, oplus, otimes
				if((len == 5) && (strncmp("omega", buffer, len) == 0)){
					newElem->SetString("w");
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}
				
				if((len == 5) && (strncmp("oplus", buffer, len) == 0)){
					newElem->SetString((char)0xC5);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				if((len == 6) && (strncmp("otimes", buffer, len) == 0)){
					newElem->SetString((char)0xC4);
					newElem->SetFontIndex(PRESTO_FONT_SYMBOL);
					return len;
				}

				return -1;
				break;
			case 'f': // forall, frac{}{}
				if((len == 6) && (strncmp("forall", buffer, len) == 0)){
					newElem->SetString((char)0x22);
					newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
					return len;
				}

				if((len == 4) && (strncmp("frac", buffer, len) == 0)){
					int tmp;

					// How many characters were read (including the '{' and '}')?
					// Note that we only pass this function the part of the text
					// buffer that contains the "{ ... }".
					tmp = Fraction(newElem, buffer + len);
					
					if(tmp < 0)
						return -1;

					return len + tmp;
				}

				return -1;
				break;
			};

	// Special characters
	switch(buffer[0]){
		case '}':
			if((len == 1) && (strncmp("}", buffer, len) == 0)){
				newElem->SetString("}");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
		case '{':
			if((len == 1) && (strncmp("{", buffer, len) == 0)){
				newElem->SetString("{");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
		case '$':
			if((len == 1) && (strncmp("$", buffer, len) == 0)){
				newElem->SetString("$");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
		case '_':
			if((len == 1) && (strncmp("_", buffer, len) == 0)){
				newElem->SetString("_");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
		case '&':
			if((len == 1) && (strncmp("&", buffer, len) == 0)){
				newElem->SetString("&");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
		case '%':
			if((len == 1) && (strncmp("%", buffer, len) == 0)){
				newElem->SetString("%");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
		case '#':
			if((len == 1) && (strncmp("#", buffer, len) == 0)){
				newElem->SetString("#");
				newElem->SetFontIndex(PRESTO_FONT_DEFAULT);
				return len;
			}

			return -1;
			break;
	};

	return -1;
}

// Implement the LaTeX commands that stack one string (m_string) on 
// top of another (buffer) and take a single argument 
// (i.e. \hat{a}, \vec{a}, etc.).
// Note that we only pass this function the part of the text
// buffer that conatins the "{ ... }".
int TextLabel::StackString(TextElem *newElem, char *buffer, CString m_string, 
						   int m_font, float m_scale)
{
	// We require an argument!
	if(buffer[0] != '{')
		return -1;

	// Skip over the '{' character.
	buffer = buffer + 1;

	char *tmp = MatchCurlyBrace(buffer);

	if(tmp == NULL)
		return -1;

	// Terminate the string prior to parsing
	tmp[0] = '\0';

	// Save the length of the string for the return value.
	// Make sure to include the '{' and '}' characters when computing the
	// string length.
	int len  = strlen(buffer) + 2;

	TextElem *tmpElem;

	// Parse the argument string (bottom) and save a pointer to it
	tmpElem = StringToLatex(buffer);

	if(tmpElem == NULL)
		return -1;

	// Copy the parsed results of the bottom string into the memory
	// provided by the caller
	*newElem = *tmpElem;

	// Free the temporary memory allocated by new
	delete tmpElem;

	// Parse the top string
	tmpElem = new TextElem(m_string, m_font, (int)(m_scale*pt_size), angle, color);

	newElem->Push(tmpElem, TEXT_TOP);

	return len;
}

// Implement the LaTeX commands that stack one string on 
// top of another and take two arguments
// (i.e. \stackrel{a}{b}). 
// Note that we only pass this function the part of the text
// buffer that conatins the "{ ... }{ ... }".
int TextLabel::StackString2(TextElem *newElem, char *buffer, float m_scale)
{
	// We require an argument!
	if(buffer[0] != '{')
		return -1;

	// Put argument one on top of argument two.
	char *arg1, *arg2, *tmp;

	// Skip over the '{' character.
	arg1 = buffer + 1;

	tmp = MatchCurlyBrace(arg1);

	if(tmp == NULL)
		return -1;

	// Terminate the string prior to parsing
	tmp[0] = '\0';

	if(tmp[1] != '{')
		return -1;

	arg2 = tmp + 2;

	tmp = MatchCurlyBrace(arg2);

	if(tmp == NULL)
		return -1;

	// Terminate the string prior to parsing
	tmp[0] = '\0';

	// Save the length of the string for the return value.
	// Make sure to include the '{' and '}' characters when computing the
	// string length.
	int len = strlen(arg1) + strlen(arg2) + 4;

#ifdef OLD_VERSION
	// Center all text about the argument
	newElem->SetOptions(TEXT_CENTER_BOTTOM);

	// Parse the first (top) argument string and save a pointer to it.
	TextElem *tmpElem = StringToLatex(arg1);

	if(tmpElem == NULL)
		return -1;

	// The top argument is scaled by an amount m_scale.
	newElem->Push(tmpElem, TEXT_TOP, m_scale);

	//tmpElem = new TextElem(m_string, m_font, (int)(m_scale*pt_size), angle, color);
	tmpElem = StringToLatex(arg2);

	newElem->Push(tmpElem, TEXT_BOTTOM);
#endif // OLD_VERSION

	TextElem *tmpElem;

	// Parse the bottom string
	tmpElem = StringToLatex(arg2);

	// Copy the parsed results of the bottom string
	*newElem = *tmpElem;

	// Free the temporary memory allocated by StringToLatex
	delete tmpElem;
	
	// Parse the first (top) argument string and save a pointer to it.
	tmpElem = StringToLatex(arg1);

	if(tmpElem == NULL)
		return -1;

	// The top argument is scaled by an amount m_scale.
	newElem->Push(tmpElem, TEXT_TOP, m_scale);

	return len;
}

// Implement the LaTeX \frac{}{} command.
// Note that we only pass this function the part of the text
// buffer that conatins the "{ ... }{ ... }".
int TextLabel::Fraction(TextElem *newElem, char *buffer)
{
	// We require an argument!
	if(buffer[0] != '{')
		return -1;

	char *arg1, *arg2, *tmp;

	// Skip over the '{' character.
	arg1 = buffer + 1;

	tmp = MatchCurlyBrace(arg1);

	if(tmp == NULL)
		return -1;

	// Terminate the string prior to parsing
	tmp[0] = '\0';

	if(tmp[1] != '{')
		return -1;

	arg2 = tmp + 2;

	tmp = MatchCurlyBrace(arg2);

	if(tmp == NULL)
		return -1;

	// Terminate the string prior to parsing
	tmp[0] = '\0';

	// Save the length of the string for the return value.
	// Make sure to include the '{' and '}' characters when computing the
	// string length.
	int len = strlen(arg1) + strlen(arg2) + 4;

	// Draw a dividing line and center all text about 
	// the dividing line
	newElem->SetOptions(TEXT_CENTER_THIS | TEXT_FRAC_LINE | TEXT_ALIGN_MID);

	// Parse the first (top) argument string and save a pointer to it.
	TextElem *tmpElem = StringToLatex(arg1);

	if(tmpElem == NULL)
		return -1;

	newElem->Push(tmpElem, TEXT_TOP, 0.8f);

	tmpElem = StringToLatex(arg2);

	newElem->Push(tmpElem, TEXT_BOTTOM, 0.8f);

	return len;
}

// Helper functions
BOOL GetTokenList(char *buffer, list<TextToken> &token_list)
{
	if(buffer == NULL)
		throw Error("Bad buffer", __FILE__, __LINE__);
	
	if(buffer[0] == '\0')
		return TRUE;

	TextToken tmp;

	if(buffer[0] == TEXT_DELIM){
		// Terminate the string
		buffer[0] = '\0';

		buffer = buffer + 1;

		if(buffer[0] == '\0')
			return TRUE;
	
		// This string should be parsed as latex.
		tmp.latex = TRUE;
	}
	else{
		// This string should NOT be parsed as latex.
		tmp.latex = FALSE;
	}

	tmp.token = buffer;

	token_list.Push(tmp);

	char *ptr = buffer;

	while(buffer){
		ptr = strchr(ptr, TEXT_DELIM);

		if(ptr == NULL)
			// This string should NOT be parsed as latex.
			return TRUE;
		
		// We found a TEXT_DELIM character
		// Was the previous character a TEXT_ESCAPE character?
		if(*(ptr-1) != TEXT_ESCAPE){
			if(tmp.latex){
				// Don't include the TEXT_DELIM character (to read 
				// next token as normal text).
				ptr[0] = '\0'; // <-- Dump the TEXT_DELIM character
				return GetTokenList(ptr + 1, token_list);
			}
			else // tmp.latex = FALSE
				// Include the TEXT_DELIM character (to read next
				// token a Latex).
				return GetTokenList(ptr, token_list);
		}
		else{
			// The TEXT_DELIM is preceeded by an escape character
			// (the user has requested a literal TEXT_DELIM).
			ptr = ptr + 1;
		}
	}

	// If we get here, there's been an error
	return FALSE;
}

// Find the matching curly brace (i.e. '}'). 
// Return null if a matching brace is not found.
char *MatchCurlyBrace(char *buffer)
{
	int i = 0;
	int count = 1;
	char c;

	// Have we read an escape character?
	int escape = FALSE;

	while((c = buffer[i]) != '\0'){
		if(!escape){
			if(c == '{'){
				count ++;
			}
			else{
				 if(c == '}'){
					 count --;
				 }
			}
		}
		
		if(c == '\\')
			escape = TRUE;
		else
			escape = FALSE;

		// We've matched up '{' and '}'.
		if(count == 0)
			return buffer + i;

		i++;
	}

	return NULL;
}

FIXED FloatToFixed(float f)
{
	long l;

	l=(long)(f*65536.0);

	return *(FIXED *)&l;
}