//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

#ifndef __PRESTO_TEXT
#define __PRESTO_TEXT

#include <math.h>

// The allowed text positions
enum {TEXT_TOP = 0,		// On top
	  TEXT_SUPER_R,		// Upper right
	  TEXT_RIGHT,		// Directly to the right 
	  TEXT_SUB_R,		// Lower right
	  TEXT_BOTTOM,		// Below
	  TEXT_LOC_NUM};	// Not a location -- the NUMBER of locations!

//#define		TOP_GAP			1
//#define		BOTTOM_GAP		1
#define		VERTICAL_GAP		1
#define		LEFT_GAP			2
#define		RIGHT_GAP			2

// The fractional height to offset superscript and subscript characters
//#define	SUPER_OFF		(0.5f)
//#define	SUB_OFF			(0.25f)

// The fractional font size (relative to the parent character) for superscript
// and supscript characters.
#define	SUPER_SUB_SCALE		(0.75f)

enum {PRESTO_FONT_DEFAULT = 0,	// Font for writing plain text
	  PRESTO_FONT_MATH,			// Font for writing text in math (latex) mode
	  PRESTO_FONT_SYMBOL,		// Font for writing symbols in math mode
	  PRESTO_FONT_SYMBOL_EXT1,	// Font for writing more symbols in math mode
	  PREST_FONT_COUNT};		// Number of fonts to load (not a font).

// The equation/symbol delimiter
#define		TEXT_DELIM		'$'
#define		TEXT_ESCAPE		'\\'

// ************************* Text alignment *************************
// The default text alignment must be modified when mixing different
// fonts (like symbol and arial) to insure that all characters are 
// correctly aligned.
//#define	PRESTO_TEXT_ALIGNMENT	(TA_BOTTOM | TA_LEFT | TA_NOUPDATECP)
#define	PRESTO_TEXT_ALIGNMENT	(TA_BASELINE | TA_LEFT | TA_NOUPDATECP)

// Space required for the dividing line in the frac{}{} command
#define		FRAC_LINE_WIDTH		1

//////////////////////////////////////////////////////////////////////
// Text element options:
//
// How should an element be centered (i.e. about the top or bottom
// childeren or current element)?
#define TEXT_CENTER_TOP			1
#define	TEXT_CENTER_THIS		2
#define	TEXT_CENTER_BOTTOM		4

// Draw a line of lenght equal to the max width of
// the top and bottom elements (i.e. \frac{}{}).
#define	TEXT_FRAC_LINE			8

// This element is aligned with it's midpoint along the baseline
#define	TEXT_ALIGN_MID			16

// Superscripts for this element go on top (like \sum)
#define	TEXT_SUPER_TOP			32

// Subscripts for this element go on bottom (like \lim)
#define	TEXT_SUB_BOTTOM			64
//////////////////////////////////////////////////////////////////////

// Track the update (or modification) state of the element to reduce
// the number of GDI calls.
#define	TEXT_UPDATE_ALL				0
#define	TEXT_SET_EXTENT				1
#define	TEXT_SET_HEIGHT				2
#define	TEXT_SET_AVE_HEIGHT			4

//////////////////////////////////////////////////////////////////////
class TextElem{
private:
	TextElem *next[5];

	static int printScaleMul;
	static int printScaleDiv;

	// Font size selected by the user
	int pt_size;

	// Font color
	COLORREF color;

	// Index into a font array
	int font;
	
	CFont *curr_font;

	// Font orientation (in degrees)
	int angle;

	// How should this element be centered (i.e.
	// centerTop, centerThis or centerBottom). Should
	// a dividing line be drawn? Etc.
	int options;

	// String to be output
	CString elem;

	// The cos and sin of the angle of rotation
	float c, s;

	// Has the element been modified? What needs to be calculated?
	int update_status;

	// Text dimentions (whose update is under the control of update_status)
	GLYPHMETRICS gm;
	CSize elem_extent;
	int elem_height;
	int offset;
	int ave_height; // The average height of a glyph (in screen dimensions).
					// Computed via the heuristic: -MulDiv(lfHeight, 1, 3).

	CFont* SetFont(CDC *pDC, LOGFONT *m_font_table);
	
	void Flush();

	// Force this element and all of its childeren to be
	// updated.
	void UpdateAll()
	{
		update_status = TEXT_UPDATE_ALL;

		for(int i = 0;i < TEXT_LOC_NUM;i++)
			if(next[i])
				next[i]->UpdateAll();
	};

public:
	TextElem *GetNext(int m_loc);

	TextElem();
	TextElem(CString &m_text, int m_font, int m_size, 
		int m_angle = 0, COLORREF m_color = 0 /* Black */,
		int m_opt = TEXT_CENTER_THIS);

	~TextElem();

	inline void SetPointSize(int m_pt_size)
	{
		pt_size = m_pt_size;
		//ave_height = pt_size/2;

		UpdateAll();
	};

	inline void SetColor(COLORREF m_color)
	{
		color = m_color;
	};

	// This function MUST be used to set the angle variable!
	// (don't set it by hand without setting 'c' and 's').
	inline void SetAngle(int m_angle)
	{
		angle = m_angle;

		// Map xy into the rotated coordinate frame:
		//     |  c  s | | x |  =  | x_p | 
		//     | -s  c | | y |     | y_p |
		// Compute the angle in radians (convert from 10*degrees).
		float theta = angle*1.74532925199e-3f;
		c = (float)(cos(theta));
		s = (float)(sin(theta));

		UpdateAll();
	};
	
	inline void SetFontIndex(int m_font)
	{
		font = m_font;

		UpdateAll();
	};

	inline void SetString(CString &m_elem)
	{
		elem = m_elem;

		UpdateAll();
	};

	inline void SetString(char *m_elem)
	{
		elem.Format("%s", m_elem);

		UpdateAll();
	};

	inline void SetString(char *m_elem, int m_len)
	{
		elem = CString(m_elem, m_len);

		UpdateAll();
	};

	// For setting single character strings
	inline void SetString(char m_c)
	{
		elem.Format("%c", m_c);

		UpdateAll();
	};

	inline void SetOptions(int m_opt)
	{
		options = m_opt;

		UpdateAll();
	};

	inline void AddOptions(int m_opt)
	{
		options |= m_opt;

		UpdateAll();
	};

	inline int GetOptions()
	{
		return options;
	};

	inline void SetPrintScaling(int m_printScaleMul, int m_printScaleDiv)
	{
		printScaleMul = m_printScaleMul;
		printScaleDiv = m_printScaleDiv;

		UpdateAll();
	};

	inline void ScalePointSize(float font_scale)
	{
		pt_size = (int)(pt_size*font_scale);

		for(int i = 0;i < TEXT_LOC_NUM;i++)
			if(next[i])
				next[i]->ScalePointSize(font_scale);

		UpdateAll();
	}

	void Copy(const TextElem &m_te);

	void Push(TextElem *pText, int loc, float font_scale = 1.0f);

	CRect GetExtent(CDC *m_pDC, LOGFONT *m_font_table,
		int m_x = 0, int m_y = 0);

	// Provide an accurate height for the elem string with the 
	// currently selected font.
	int GetElemHeight(CDC *m_pDC, LOGFONT *m_font_table, int *offset = NULL);

	CSize GetTextExtent(CDC *m_pDC);

	void Draw(CDC *m_pDC, LOGFONT *m_font_table,
		int m_x, int m_y);
};

// A wrapper for TextElem
class TextLabel{
private:
	TextElem *label;
	LOGFONT *font_table;
	int num_font;

	// The text angle in degrees * 10.
	int angle;

	// The size in points
	int pt_size;

	COLORREF color;

	TextElem* StringToLatex(char *buffer);
	int GetEscape(TextElem *newElem, char *buffer, int len);
	BOOL Parse(CString &m_text);

	// Implement the LaTeX commands that stack one string (m_string) on 
	// top of another (buffer) and take a single argument 
	// (i.e. \hat{a}, \vec{a}, etc.).
	int StackString(TextElem *newElem, char *buffer, CString m_string, 
					int m_font, float m_scale = 1.0f);

	// Implement the LaTeX commands that stack one string on 
	// top of another and take two arguments
	// (i.e. \stackrel{a}{b}).
	int StackString2(TextElem *newElem, char *buffer, float m_scale = 1.0f);

	// Implement the LaTeX \frac{}{} command.
	// Note that we only pass this function the part of the text
	// buffer that contains the "{ ... }{ ... }".
	int Fraction(TextElem *newElem, char *buffer);
public:
	TextLabel();
	~TextLabel();

	void SetLabel(CString &m_text, LOGFONT &m_font, COLORREF m_color = 0);
	void Draw(CDC *m_pDC, int m_x, int m_y);
	CRect GetExtent(CDC *m_pDC);

	inline BOOL IsSet()
	{
		return (label != NULL);
	};

	void Copy(const TextLabel& m_tl);

	TextLabel &operator=(const TextLabel& m_tl)
	{
		Copy(m_tl);

		return (*this);
	};

	inline void SetPrintScaling(int m_printScaleMul, int m_printScaleDiv)
	{
		if(label)
			label->SetPrintScaling(m_printScaleMul, m_printScaleDiv);
	};
};

#endif // __PRESTO_TEXT