//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
#include "EquationUtil.h"
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

void EqElemVal::Set(list<EqToken> &m_token, EqElem *m_p, list<EqToken> &varstack)
{
	// Set the parent pointer
	P = m_p;

	if(m_token[0].token_type == VALUE_CONST){
		isConst = TRUE;
		constval = m_token[0].numeric_val;
	}
	else{
		index = AddToStack(m_token[0], varstack);
	}

	m_token.Remove(0);
}

void EqElemVal::Push(list<EqToken> &m_token, list<EqToken> &varstack)
{
	if(m_token.Length() == 0)
		// Nothing to do!
		return;

	if(P == NULL){
		// Allocate space for the parent element (must be a binary op)
		if(!isBinaryOp(m_token[0].token_type))
			throw "Parse error: binary operator expected after value";

		P = new EqElemBinary;

		BINARY(P)->L = this;

		P->Set(m_token, NULL, varstack);
		P->Push(m_token, varstack);
		return;
	}

	//////////////////////////////////////////////////////////////////////
	// Make sure proper order of operations is maintained between binary
	// operators (i.e. ^,/,*,+,-).
	//////////////////////////////////////////////////////////////////////
	if(isBinaryOp(P->GetType())){
		if((P->GetType() == BINARY_ADD) || (P->GetType() == BINARY_SUB)){
			if((m_token[0].token_type == BINARY_DIV) || 
			   (m_token[0].token_type == BINARY_MULTI) ||
			   (m_token[0].token_type == BINARY_POW)){
				
				// Insert the new operator
				EqElem *tmp_P = P;

				P = new EqElemBinary;
				
				// Attach the old parent to this new element
				BINARY(tmp_P)->R = P;

				// Attach the new parent to this element
				BINARY(P)->L = this;

				P->Set(m_token, tmp_P, varstack);
				P->Push(m_token, varstack);
				return;
			}
		}

		if((P->GetType() == BINARY_DIV) || (P->GetType() == BINARY_MULTI)){
			if(m_token[0].token_type == BINARY_POW){
				
				// Insert the new operator
				EqElem *tmp_P = P;

				P = new EqElemBinary;
				
				// Attach the old parent to this new element
				BINARY(tmp_P)->R = P;

				// Attach the new parent to this element
				BINARY(P)->L = this;

				P->Set(m_token, tmp_P, varstack);
				P->Push(m_token, varstack);
				return;
			}
		}

		P->Push(m_token, varstack);
		return;
	}
	else{ // Unary operator

		// Is the parent a unary brace or a unary negative sign?
		if((P->GetType() == UNARY_PAREN) || 
		  ((P->GetType() == UNARY_NEG) &&
		  ((m_token[0].token_type == BINARY_DIV) || 
		   (m_token[0].token_type == BINARY_MULTI) ||
		   (m_token[0].token_type == BINARY_POW)))){
			if(!isBinaryOp(m_token[0].token_type))
				throw "Missing a binary operator between braces or after '-'";

			// The brace operator must get special treatment
			// Insert the new operator
			EqElem *tmp_P = P;

			P = new EqElemBinary;

			// Attach the old parent to this new element
			UNARY(tmp_P)->C = P;

			// Make the current element the left operatand
			BINARY(P)->L = this;

			P->Set(m_token, tmp_P, varstack);
			P->Push(m_token, varstack);
			return;
		}

		P->Push(m_token, varstack);
	}
}

double EqElemVal::Eval(double *m_eval_stack)
{
	// Don't test the stack pointer BEFORE evaluating constants!
	// stack == NULL is a valid stack for a constant value!
	if(isConst)
		return constval;

	if(safe_mode){
		if(m_eval_stack == NULL)
			throw "NULL stack";

		if(index < 0)
			throw "stack underflow Eval()";

		//if(index >= varstack.Length())
		//	throw "stack overflow Eval()";
	}

	return m_eval_stack[index];
}

#ifdef _DEBUG
void EqElemVal::DebugEval(FILE *fout, list<EqToken> &varstack)
{
	if(isConst){
		fprintf(fout, "%g", constval);
		return;
	}

	if(safe_mode){
		if(index < 0)
			throw "stack underflow Eval()";

		if(index >= varstack.Length())
			throw "stack overflow Eval()";
	}
	
	//fprintf(fout, "(%s)", varstack[index].id);
	fprintf(fout, "%s", varstack[index].id);
}
#endif // _DEBUG

int EqElemVal::AddToStack(EqToken &m_tmp, list<EqToken> &varstack)
{
	for(int i = 0;i < varstack.Length();i++){
		// Is the element already in the stack?
		if(strcmp(varstack[i].id, m_tmp.id) == 0)
			return i;
	}

	// The element is NOT in the stack
	varstack.Push(m_tmp);

	return varstack.Length() - 1;
}

EqElem* EqElemVal::Copy(EqElem *m_p)
{
	EqElem* tmp = new EqElemVal;

	VALUE(tmp)->P = m_p;

	VALUE(tmp)->index = index;
	VALUE(tmp)->isConst = isConst;
	VALUE(tmp)->constval = constval;

	return tmp;
}

BOOL EqElemVal::IsEqual(EqElem *m_elem)
{
#ifdef _DEBUG
	if(m_elem == NULL)
		throw "NULL pointer in EqElemVal::IsEqual";
#endif // _DEBUG

	if(m_elem->GetType() != GetType()){
		return FALSE;
	}

	if(isConst && VALUE(m_elem)->isConst){
		if(fabs(constval - VALUE(m_elem)->constval) < EQ_EPSILON)
			return TRUE;
	}
	else{
		if(!isConst && !VALUE(m_elem)->isConst){
			return index == VALUE(m_elem)->index;
		}
	}

	return FALSE;
}

EqElem* EqElemVal::Deriv(int m_var_index, EqElem *m_p)
{
	EqElem* tmp = new EqElemVal;
	
	tmp->P = m_p;
	VALUE(tmp)->isConst = TRUE;

	// If const, then the derivative is zero.
	if(isConst || (index != m_var_index)){	
		VALUE(tmp)->constval = 0.0;
	}
	else{ // The derivative is one.
		VALUE(tmp)->constval = 1.0;
	}

	return tmp;
}

void EqElemVal::CleanUp()
{
	// Do nothing!
}
