//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
// A class for dynamically evaluating text-based mathematical 
// expressions. 10/19/01 J. D. Gans
//
// ** Change Log **
//
//	12/18/01
//		Added support for string constants like "Pi" (= 3.1415 ...).
//	9/25/02
//		Added support for functions with up to three independant 
//		variables. More variables can be supported by overloading
//		the DefineIndepVar() and operator () functions. The use of
//		overloaded functions avoids problems with variable length
//		argument lists.
//

#ifndef __EQUATION
#define	__EQUATION

#include <stdlib.h>
#include "../list.h"

#ifdef MEMORY_DEBUG
#include "memory.h"
#endif

#ifndef _WINDOWS
typedef int BOOL;
#endif 

#ifndef TRUE
#define	TRUE	1
#endif

#ifndef FALSE
#define	FALSE	0
#endif

// A small number
#define		EQ_EPSILON		1.0e-14

// The current version of this class
#define		EQ_VERSION		1.0f

// Enumerate the allowed operations
enum {BINARY_ADD = 0, 
	  BINARY_SUB, 
	  BINARY_DIV, 
	  BINARY_MULTI, 
	  BINARY_POW,
	  UNARY_NEG,
	  UNARY_PLUS,
	  UNARY_PAREN,
	  UNARY_L_BRACE,
	  UNARY_R_BRACE,
	  UNARY_SQRT, 
	  UNARY_COS, 
	  UNARY_SIN, 
	  UNARY_TAN, 
	  UNARY_ACOS,
	  UNARY_ASIN,
	  UNARY_ATAN,
	  UNARY_SEC, 
	  UNARY_CSC, 
	  UNARY_COT,
	  UNARY_COSH,
	  UNARY_SINH,
	  UNARY_TANH,
	  UNARY_EXP,
	  UNARY_LOG,
	  UNARY_LOG10,
	  UNARY_ABS,
	  VALUE_CONST,
	  VALUE_VAR};

class EqToken{
public:
	int token_type;
	
	// For tokens of type VALUE_VAR:
	char *id;

	// For tokens of type VALUE_CONST:
	double numeric_val;

	EqToken()
	{
		id = NULL;
	};

	~EqToken()
	{
		if(id)
			delete [] id;
	};

	EqToken& operator = (const EqToken& m_copy);
	
	char *Set(char *m_input, int len);

	int GetOperator(char *m_input, int len);
	BOOL isNumber(char *m_input, int len);
	BOOL isVariable(char *m_input, int len);
};

struct EqElem{
	// The parent element
	EqElem *P; 
	
	static int safe_mode;

	virtual double Eval(double *m_eval_stack) = 0;

#ifdef _DEBUG
	virtual void DebugEval(FILE *fout, list<EqToken> &varstack) = 0;
#endif // _DEBUG

	virtual int GetType() = 0;
	virtual void Push(list<EqToken> &m_token, list<EqToken> &varstack) = 0;
	virtual void Set(list<EqToken> &m_token, EqElem *m_p, 
		list<EqToken> &varstack) = 0;

	virtual EqElem* Copy(EqElem *m_p) = 0;
	virtual BOOL IsEqual(EqElem *m_elem) = 0;
	virtual EqElem* Deriv(int m_var_index, EqElem *m_p) = 0;

	virtual void CleanUp() = 0;
};

struct EqElemBinary: public EqElem{
	// The childeren
	EqElem *L, *R;

	// What binary operation are we performing?
	int binary_op;
	
	void Push(list<EqToken> &m_token, list<EqToken> &varstack);
	void Set(list<EqToken> &m_token, EqElem *m_p, list<EqToken> &varstack);

	int GetType()
	{
		return binary_op;
	};

	inline double Eval(double *m_eval_stack);

#ifdef _DEBUG
	void DebugEval(FILE *fout, list<EqToken> &varstack);
#endif // _DEBUG

	EqElem* Copy(EqElem *m_p);
	BOOL IsEqual(EqElem *m_elem);
	EqElem* Deriv(int m_var_index, EqElem *m_p);

	EqElemBinary(EqElem* m_p = NULL)
	{
		P = m_p;
		L = R = NULL;
	};

	EqElemBinary(int m_op)
	{
		binary_op = m_op;
		P = L = R = NULL;
	};

	void CleanUp();
};

struct EqElemUnary: public EqElem{
	// The single child
	EqElem *C;

	// What unary operation are we performing?
	int unary_op;

	void Push(list<EqToken> &m_token, list<EqToken> &varstack);
	void Set(list<EqToken> &m_token, EqElem *m_p, list<EqToken> &varstack);

	int GetType()
	{
		return unary_op;
	};

	inline double Eval(double *m_eval_stack);

#ifdef _DEBUG
	void DebugEval(FILE *fout, list<EqToken> &varstack);
#endif // _DEBUG

	inline double sec(double m_x);
	inline double csc(double m_x);
	inline double cot(double m_x);

	EqElem* Copy(EqElem *m_p);
	BOOL IsEqual(EqElem *m_elem);
	EqElem* Deriv(int m_var_index, EqElem *m_p);

	EqElemUnary(EqElem* m_p = NULL)
	{
		P = m_p;
		C = NULL;
	};

	EqElemUnary(int m_op)
	{
		unary_op = m_op;
		P = C = NULL;
	};

	void CleanUp();
};

struct EqElemVal: public EqElem{
	// A index into a variable stack
	int index;

	BOOL isConst;
	double constval;

	int AddToStack(EqToken &m_tmp, list<EqToken> &varstack);

	void Push(list<EqToken> &m_token, list<EqToken> &varstack);
	void Set(list<EqToken> &m_token, EqElem *m_p, list<EqToken> &varstack);

	int GetType()
	{
		if(isConst)
			return VALUE_CONST;

		return VALUE_VAR;
	};

	inline double Eval(double *m_eval_stack);

#ifdef _DEBUG
	void DebugEval(FILE *fout, list<EqToken> &varstack);
#endif // _DEBUG

	EqElem* Copy(EqElem *m_p);
	BOOL IsEqual(EqElem *m_elem);
	EqElem* Deriv(int m_var_index, EqElem *m_p);

	EqElemVal(EqElem* m_p = NULL)
	{
		P = m_p;
		isConst = FALSE;
	};

	void CleanUp();
};

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class SEquation{
private:
	
	// Version control for disk IO
	float version;

	EqElem *tree;
	char *last_error;
	char *input_string;

	double *eval_stack;
	BOOL *stack_valid;

	list<int> indep_var;

	list<EqToken> varstack;

	char *Tokenize(char *m_input, list<EqToken> &tokenlist);
	void MakeAST(list<EqToken> &tokenlist);

	// Tree operations
	void RemoveUnaryElem(EqElemUnary *ptr);
	void InsertUnaryParent(EqElem *ptr, EqElemUnary *ptr_P);

	// Remove brace operators
	void SimplifyBrace(EqElem *m_ptr);

	// Remove constant expressions
	BOOL SimplifyConst(EqElem *m_ptr);

	// Remove zeros
	BOOL SimplifyZero(EqElem *m_ptr);

	// Remove ones
	BOOL SimplifyOne(EqElem *m_ptr);

	// Remove any unary addition operators
	BOOL SimplifyUnaryPlus(EqElem *m_ptr);

	// Remove any functions of inverse functions
	BOOL SimplifyInverse(EqElem *m_ptr);

	// Remove repeated unary minus signs and move unary minus signs
	// towards the root of the tree
	BOOL SimplifyUnarySub(EqElem *m_ptr);

	// Absord binary "-" into constants when possible
	BOOL SimplifyBinarySub(EqElem *m_ptr);
	
	// Remove binary operations ("*/+-") acting on equal operands
	BOOL SimplifyIdentical(EqElem *m_ptr);
public:
	SEquation();
	SEquation(const char *m_input);
	SEquation(SEquation &m_copy);

	~SEquation();
	
	void ClearError();

	char* GetError() const
	{
		return last_error;
	};

	BOOL Set(const char *m_input);
	
	// Flush the current equation
	void Reset();

	// Assign a numeric value to a symbolic constant
	void SetVar(int num, float m_value);

	// Return the numeric value of a symbolic constant
	float GetVar(int num);

	int NumVar();

	int NumIndepVar()
	{
		return indep_var.Length();
	};

	int GetIndepVar(int i)
	{
		if((i < 0) || (i >= indep_var.Length()))
			return -1;

		return indep_var[i];
	};

	BOOL IsIndepVar(int m_i)
	{
		for(int i = 0;i < indep_var.Length();i++)
			if(m_i == indep_var[i])
				return TRUE;

		return FALSE;
	};

	char *GetVarName(int num);

	// Define the arguments of the equation. Since the number
	// of arguments is not known in advance, several overloaded
	// operators are provided. This is a limitation (since a large
	// number of independant variables will require a new function
	// and a recompile) it avoids the nastyness variable length 
	// argument lists.
	BOOL DefineIndepVar();
	BOOL DefineIndepVar(char *m_one);
	BOOL DefineIndepVar(char *m_one, char *m_two);
	BOOL DefineIndepVar(char *m_one, char *m_two, char *m_three);
		
	// Is the equation ready to be evaluated?
	// If m_test_stack is set to TRUE, IsValid checks to see if
	// the equation is ready to be numerically evaluated. If set to
	// FALSE, IsValid does not test for numerical evaluation (and
	// symbolic constants are OK).
	BOOL IsValid(BOOL m_test_stack = TRUE);

	float operator () ();
	float operator () (float m_one);
	float operator () (float m_one, float m_two);
	float operator () (float m_one, float m_two, float m_three);

#ifdef _DEBUG
	void operator () (FILE *fout);
#endif // _DEBUG

	// Attempt to simplify a valid equation
	void Simplify();

	// Compute the derivative of the current equation
	SEquation Deriv(char *m_var);
	SEquation Deriv(int m_var_index);

	// Allow equations to be copied
	SEquation& operator =(SEquation &copy);
};

class EqElemIterator{
private:
	EqElem *root, *cur;

	// What elements have we restricted the search to?
	int mask;

	list< EqElem* > ptr_stack;

	BOOL TestMask();
public:
	// The constructor (note that no destructor is needed).
	EqElemIterator(EqElem* m_root, int m_mask = -1);

	// Return the next element in the tree (or NULL if we reach the end of
	// the tree or run out of elements consistent with the mask).
	EqElem* operator () ();
};

#endif // __EQUATION
