//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
#include "EquationUtil.h"
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

// Attempt to simplify a valid equation.
// Simplification involves the following steps:
//		(1) Remove all "()" operators by contraction of tree elements
//		(1a) Remove all unary "+" operators by contraction of tree elements
//		(1b) Move unary "-" operators up the tree or absorb into constants
//			a*-b = -(a*b),
//			-a/b = -(a/b),
//			a*-c = a*(-c),
//			-a*c = a*(-c),
//		(1c) Remove unary "-" operators by the following contraction rules
//			--a = a
//			a+-b = a - b
//		(1d) Remove binary "-" operators by the following contraction rules
//			a - c = a + (-c)
//			c - a = -((-c) + a)
//		(2) Evaluate all constant subexpressions, i.e.
//				c1 - c2,
//				c1 + c2,
//				c1 * c2,
//				c1 / c2,
//				c1 ^ c2,
//				unary(c1), <-- This includes "+" and "-".
//			where c1 and c2 are constant values.
//		(2a) Perform a contraction of tree elements with equal
//			recedence. For instance:
//				(a + c1) + c2 = a + (c1 + c2),
//				(a + c1) - c2 = a + (c1 - c2),
//				(a - c1) + c2 = a + (c2 - c1),
//				(a * c1) / c2 = a * (c1 / c2),
//				(a / c1) * c2 = a * (c2 / c1),
//			where a is a variable and c1 and c2 are constants. This simplification
//			is non-local within the AST, i.e.
//				a * c1 * b / d / e * c2 = a * (c1 * c2)  * b / d / e,
//			where b, d and e are all variables.
//		(3a) Reduce inverse functions, i.e.
//			exp(ln(a)) = a,
//			ln(exp(a)) = a,
//		(3b) Power reduction. This includes
//				a^b * a^d = a^(b + d),
//				a^b / a^d = a^(b - d)
//				The following identities will assist simplification
//				1/a = a^(-1),
//				a = a^1.
//		(4) Remove all constant "zeros" via the following rules
//				a + 0 = a,
//				a - 0 = a,
//				a * 0 = a,
//				a ^ 0 = 1,
//				0 ^ a = 0,
//			where a is a variable.
//		(5) Remove all constant "ones" via the following rules
//				a*1 = a,
//				a^1 = a,
//				1^a = 1,
//				a/1 = 1,
//			where a is a variable.
//		(6) Reduce identical expressions:
//			(a + b)/(a + b) = 1
//			(a + b)*(a + b) = (a + b)^2
//			(a + b)+(a + b) = (a+b)*2
//			(a + b)-(a + b) = 0

void SEquation::Simplify()
{
	// Only valid equations can be simplified
	if(!IsValid(FALSE))
		return;

	try{
		BOOL reduce = TRUE;

		// Keep looping through reduction routines as long
		// as the equation continues to be simplified
		while(reduce){
			reduce = FALSE;

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG

			SimplifyBrace(tree);

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG

			while(SimplifyUnaryPlus(tree))
				reduce = TRUE;

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG


			while(SimplifyUnarySub(tree))
				reduce = TRUE;

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG

			while(SimplifyBinarySub(tree))
				reduce = TRUE;
			
#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG

			while(SimplifyConst(tree)){
				reduce = TRUE;

#ifdef _DEBUG
				CheckTree(tree);
#endif // _DEBUG

				while(SimplifyZero(tree))
					reduce = TRUE;
#ifdef _DEBUG
				CheckTree(tree);
#endif // _DEBUG

				while(SimplifyOne(tree))
					reduce = TRUE;
#ifdef _DEBUG
				CheckTree(tree);
#endif // _DEBUG
			}

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG
			
			while(SimplifyInverse(tree))
				reduce =  TRUE;
#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG


			while(SimplifyZero(tree))
				reduce = TRUE;
#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG

			while(SimplifyOne(tree))
				reduce = TRUE;

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG

//#ifdef NOT_READY_FOR_PRIME_TIME_YET
			while(SimplifyIdentical(tree)){
				reduce = TRUE;

#ifdef _DEBUG
				CheckTree(tree);
#endif // _DEBUG

				// Make sure we don't try and simplify zero and one!
				while(SimplifyZero(tree))
					reduce = TRUE;
#ifdef _DEBUG
				CheckTree(tree);
#endif // _DEBUG

				while(SimplifyOne(tree))
					reduce = TRUE;
#ifdef _DEBUG
				CheckTree(tree);
#endif // _DEBUG
			}

//#endif // NOT_READY_FOR_PRIME_TIME_YET

#ifdef _DEBUG
			CheckTree(tree);
#endif // _DEBUG
		}
	}
	catch(char *m_error){

		last_error = m_error;
	}
}

// Remove all "()" operators by contraction of tree elements
void SEquation::SimplifyBrace(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyBrace";
#endif // _DEBUG

	if(isUnaryOp(m_ptr->GetType())){
		if(m_ptr->GetType() == UNARY_PAREN){
			// Remove this element from the tree!
			if(m_ptr->P == NULL){
				// This is the root node!

				tree = UNARY(m_ptr)->C;
				tree->P = NULL;

				// Disconnect this node from the tree ...
				UNARY(m_ptr)->C = NULL;

				// and delete it!
				m_ptr->CleanUp();
				delete m_ptr;

				return;
			}

			// Not the root node, so we'll need to reconnect the parent
			if(isBinaryOp(m_ptr->P->GetType())){

				// Are we the left or right child?
				if(BINARY(m_ptr->P)->L == m_ptr){

					// Attach the parent
					BINARY(m_ptr->P)->L = UNARY(m_ptr)->C;

					// Attach the child
					UNARY(m_ptr)->C->P = m_ptr->P;

					// Disconnect this node
					UNARY(m_ptr)->C = NULL;

					// Delete this node
					m_ptr->CleanUp();
					delete m_ptr;
					return;
				}

				if(BINARY(m_ptr->P)->R == m_ptr){
					// Attach the parent
					BINARY(m_ptr->P)->R = UNARY(m_ptr)->C;

					// Attach the child
					UNARY(m_ptr)->C->P = m_ptr->P;

					// Disconnect this node
					UNARY(m_ptr)->C = NULL;

					// Delete this node
					m_ptr->CleanUp();
					delete m_ptr;
					return;
				}

				// If we get here, it's an error!
				throw "Disconnected tree in SimplifyBrace";
			}
			
			// Must be unary
			// Attach the parent
			UNARY(m_ptr->P)->C = UNARY(m_ptr)->C;

			// Attach the child
			UNARY(m_ptr)->C->P = m_ptr->P;

			// Disconnect this node
			UNARY(m_ptr)->C = NULL;

			// Delete this node
			m_ptr->CleanUp();
			delete m_ptr;
			return;
		}
		else{
			SimplifyBrace(UNARY(m_ptr)->C);
		}
	}
	else{
		if(isBinaryOp(m_ptr->GetType())){
			SimplifyBrace(BINARY(m_ptr)->L);
			SimplifyBrace(BINARY(m_ptr)->R);
		}
		// Must be a value, don't bother simplifying
	}
}

// Evaluate all constant subexpressions, i.e.
//	c1 - c2,
//	c1 + c2,
//	c1 * c2,
//	c1 / c2,
//	c1 ^ c2,
//	unary(c1),
//	where c1 and c2 are constant values.
// Return TRUE if a simplification has been made, FALSE
// otherwise.
BOOL SEquation::SimplifyConst(EqElem *m_ptr)
{

#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyConst";
#endif // _DEBUG

	if(isUnaryOp(m_ptr->GetType())){
		EqElemUnary *tmp = UNARY(m_ptr);

		if(tmp->C->GetType() == VALUE_CONST){
			// We have a unary operator acting on
			// a constant, simplify!
			
			// Make a new element
			EqElem *sc = new EqElemVal;

			sc->P = tmp->P;
			
			// Attach the parent, if P != NULL
			if(tmp->P != NULL){
				if(isUnaryOp(tmp->P->GetType())){
					UNARY(tmp->P)->C = sc;
				}
				else{
					// Must be binary
					if(BINARY(tmp->P)->L == tmp){
						BINARY(tmp->P)->L = sc;
					}
					else{ // ((EqElemBinary *)(tmp->P))->R == tmp
						BINARY(tmp->P)->R = sc;
					}
				}
			}
			else{
				// We will be making a new root node, so set
				// the tree pointer.
				tree = sc;
			}

			VALUE(sc)->isConst = TRUE;
			VALUE(sc)->constval = tmp->Eval(NULL);
	
			// Remove this node
			m_ptr->CleanUp();
			delete m_ptr;

			return TRUE;
		}

		return SimplifyConst(tmp->C);
	}
	else{
		if(isBinaryOp(m_ptr->GetType())){
			EqElemBinary *tmp = BINARY(m_ptr);

			if(tmp->GetType() == BINARY_POW){
				if((tmp->L->GetType() == VALUE_CONST) && (tmp->R->GetType() == VALUE_CONST)){
					EqElemVal *c_l = VALUE(tmp->L);
					EqElemVal *c_r = VALUE(tmp->R);

					c_l->constval = tmp->Eval(NULL);
					c_r->constval = 1.0;

					return TRUE;
				}
			}
			else{ // BINARY_ADD, BINARY_SUB, BINARY_DIV, BINARY_MULTI

				// Keep track of the sign or power (sop)
				int sop_l = 1, sop_r = 1;

				if((tmp->GetType() == BINARY_SUB) || (tmp->GetType() == BINARY_DIV))
					sop_r = -1;

				EqElemVal *c_l = GetNextConstVal(tmp->L, sop_l, tmp->GetType());
				EqElemVal *c_r = GetNextConstVal(tmp->R, sop_r, tmp->GetType());

				// Don't try to simplify 1, since one can be a product of
				// simplification itself and is handled separately.
				if(c_l && c_r && !isElemOne(c_l) && !isElemOne(c_r)){
					// We have a binary operator acting on
					// two constants, simplify by modifying the left constant:
					// c_l = c_l + sign_l*sign_r*c_r
					// or
					// c_l = c_l * c_r^(sign_l*sign_r)
					// and removing the right constant

					if((tmp->GetType() == BINARY_ADD) || 
					   (tmp->GetType() == BINARY_SUB)){
						c_l->constval += sop_l*sop_r*c_r->constval;
						c_r->constval = 0.0;
					}
					else{
						if(sop_l*sop_r > 0)
							c_l->constval *= c_r->constval;
						else{
							if(sop_l < 0){ // sop_r > 0
								// Make sure we don't try to divide by zero!
								// Simplification of zeros is handled separately.
								if(isElemZero(c_l) || isElemZero(c_r))
									return FALSE;

								c_l->constval = c_r->constval/c_l->constval;
							}
							else{ // sop_l > 0
								// Make sure we don't try to divide by zero!
								// Simplification of zeros is handled separately.
								if(isElemZero(c_l) || isElemZero(c_r))
									return FALSE;

								c_l->constval = c_l->constval/c_r->constval;
							}
						}
						c_r->constval = 1.0;
					}

					// Continue to simplify the left and right elements
					SimplifyConst(tmp->L);
					SimplifyConst(tmp->R);

					return TRUE;
				}
			}
			
			BOOL sl = SimplifyConst(tmp->L);
			BOOL sr = SimplifyConst(tmp->R);

			return sl || sr;
		}
	}

	return FALSE;
}

//	Remove all constant "zeros" via the following rules:
//	a + 0 = a,
//	0 + a = a,
//	a - 0 = a,
//	0 - a = -a,
//	a * 0 = 0,
//	0 * a = 0
//	0 / a = 0,
//	a ^ 0 = 1,
//	0 ^ a = 0,
//	where a is a variable. Return TRUE if a simplification has been
//	made and FALSE otherwise.
BOOL SEquation::SimplifyZero(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyZero";
#endif // _DEBUG

	if(isBinaryOp(m_ptr->GetType())){
		EqElemBinary *tmp = (EqElemBinary *)m_ptr;

		BOOL simplify = FALSE;
		BOOL simplifyElem = FALSE;

		EqElem *tmpElem = NULL;

		switch(m_ptr->GetType()){
			case BINARY_ADD:
				if(isElemZero(tmp->L)){
					//	0 + a = a
					simplify = TRUE;
					tmpElem = tmp->R;
					
					// Sever the connection between the binary operator
					// an the right element
					tmp->R = NULL;

					// Simplify this element!
					simplifyElem = TRUE;
				}
				else{
					if(isElemZero(tmp->R)){
						//	a + 0 = a
						simplify = TRUE;
						tmpElem = tmp->L;

						// Sever the connection between the binary operator
						// an the left element
						tmp->L = NULL;

						// Simplify this element!
						simplifyElem = TRUE;
					}
				}
				break;
			case BINARY_SUB:
				if(isElemZero(tmp->L)){
					// 0 - a = - a
					simplify = TRUE;
					// Handle the negation here!
					tmpElem = new EqElemUnary;

					// Connect parent to child
					UNARY(tmpElem)->C = tmp->R;

					// Connect childe to parent
					tmp->R->P = tmpElem;

					UNARY(tmpElem)->unary_op = UNARY_NEG;

					// Sever the connection between the binary operator
					// an the right element
					tmp->R = NULL;
					
					// Simplify this element!
					simplifyElem = TRUE;
				}
				else{
					if(isElemZero(tmp->R)){
						// a - 0 = a
						simplify = TRUE;
						tmpElem = tmp->L;

						// Sever the connection between the binary operator
						// an the left element
						tmp->L = NULL;

						// Simplify this element!
						simplifyElem = TRUE;
					}
				}
				break;
			case BINARY_DIV:
				if(isElemZero(tmp->L)){
					// 0 / a = 0
					simplify = TRUE;

					tmpElem = tmp->L;
					
					// Sever the connection between the binary operator
					// an the left element
					tmp->L = NULL;
				}
				break;
			case BINARY_MULTI:
				if(isElemZero(tmp->L)){
					// 0 * a = 0
					simplify = TRUE;

					tmpElem = tmp->L;
					
					// Sever the connection between the binary operator
					// an the left element
					tmp->L = NULL;
				}
				else{
					if(isElemZero(tmp->R)){
						// a * 0 = 0 
						simplify = TRUE;

						tmpElem = tmp->R;
					
						// Sever the connection between the binary operator
						// an the left element
						tmp->R = NULL;
					}
				}
				break;
			case BINARY_POW:
				if(isElemZero(tmp->L)){
					// 0 ^ a = 0
					simplify = TRUE;

					tmpElem = tmp->L;

					// Sever the connection between the binary operator
					// an the left element
					tmp->L = NULL;
				}
				else{
					if(isElemZero(tmp->R)){
						// a ^ 0 = 1
						simplify = TRUE;
						
						tmpElem = tmp->R;
						
						// Change the constant element from 0 to 1
						VALUE(tmpElem)->constval = 1.0;

						// Sever the connection between the binary operator
						// an the left element
						tmp->R = NULL;
					}
				}
				break;
		};

		if(simplify){
#ifdef _DEBUG
			if(tmpElem == NULL)
				throw "NULL tmp pointer in SimplifyZero()";
#endif // _DEBUG

			if(tmp->P == NULL){
				// This is the root node of the tree!
				tree = tmpElem;

				tree->P = NULL;

				tmp->CleanUp();
				delete tmp;
			}
			else{
				// This isn't the root node of the tree, so we'll need to connect
				// the parent element to the new child element and the child to the
				// parent
					
				tmpElem->P = tmp->P;	// Connect the child

				// Connect the parent
				if(isBinaryOp(tmp->P->GetType())){
					EqElemBinary *tmpP = BINARY(tmp->P);

					if(tmpP->L == tmp){
						tmpP->L = tmpElem;
					}
					else{ // tmpP->R == tmp
						tmpP->R = tmpElem;
					}
				}
				else{
					// Must be a unary operator
					UNARY(tmp->P)->C = tmpElem;
				}

				// Delete the element that is no longer needed
				tmp->CleanUp();
				delete tmp;
			}

			if(simplifyElem)
				SimplifyZero(tmpElem);

			return TRUE;
		}
		else{
			BOOL sl = SimplifyZero(tmp->L);
			BOOL sr = SimplifyZero(tmp->R);

			return sl || sr;
		}
	}

	if(isUnaryOp(m_ptr->GetType()))
		return SimplifyZero(((EqElemUnary *)m_ptr)->C);

	// If we get to this point, the element is a value, so do nothing!
	return FALSE;
}

//	Remove all constant "ones" via the following rules:
//	a*1 = a,
//	1*a = a,
//	a^1 = a,
//	1^a = 1,
//	a/1 = 1,
//	where a is a variable. Return TRUE if a simplification has been
//	made and FALSE otherwise.
BOOL SEquation::SimplifyOne(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyOne";
#endif // _DEBUG

	if(isBinaryOp(m_ptr->GetType())){
		EqElemBinary *tmp = BINARY(m_ptr);

		BOOL simplify = FALSE;

		EqElem *tmpElem = NULL;

		switch(m_ptr->GetType()){
			case BINARY_DIV:
				if(isElemOne(tmp->R)){
					// a / 1 = a
					simplify = TRUE;

					tmpElem = tmp->L;
					
					// Sever the connection between the binary operator
					// an the left element
					tmp->L = NULL;
				}
				break;
			case BINARY_MULTI:
				if(isElemOne(tmp->L)){
					// 1 * a = a
					simplify = TRUE;

					tmpElem = tmp->R;
					
					// Sever the connection between the binary operator
					// an the right element
					tmp->R = NULL;
				}
				else{
					if(isElemOne(tmp->R)){
						// a * 1 = a 
						simplify = TRUE;

						tmpElem = tmp->L;
					
						// Sever the connection between the binary operator
						// an the left element
						tmp->L = NULL;
					}
				}
				break;
			case BINARY_POW:
				if(isElemOne(tmp->L) || isElemOne(tmp->R)){
					// 1 ^ a = 1
					// a ^ 1 = a
					simplify = TRUE;

					tmpElem = tmp->L;

					// Sever the connection between the binary operator
					// an the left element
					tmp->L = NULL;
				}
				break;
		};

		if(simplify){
#ifdef _DEBUG
			if(tmpElem == NULL)
				throw "NULL tmp pointer in SimplifyOne()";
#endif // _DEBUG

			if(tmp->P == NULL){
				// This is the root node of the tree!
				tree = tmpElem;

				tree->P = NULL;
				
				tmp->CleanUp();
				delete tmp;
			}
			else{
				// This isn't the root node of the tree, so we'll need to connect
				// the parent element to the new child element and the child to the
				// parent
					
				tmpElem->P = tmp->P;	// Connect the child

				// Connect the parent
				if(isBinaryOp(tmp->P->GetType())){
					EqElemBinary *tmpP = BINARY(tmp->P);

					if(tmpP->L == tmp){
						tmpP->L = tmpElem;
					}
					else{ // tmpP->R == tmp
						tmpP->R = tmpElem;
					}
				}
				else{
					// Must be a unary operator
					UNARY(tmp->P)->C = tmpElem;
				}

				// Delete the element that is no longer needed
				tmp->CleanUp();
				delete tmp;
			}

			return TRUE;
		}
		else{
			BOOL sl = SimplifyOne(tmp->L);
			BOOL sr = SimplifyOne(tmp->R);

			return sl || sr;
		}
	}

	if(isUnaryOp(m_ptr->GetType()))
		return SimplifyOne(UNARY(m_ptr)->C);

	// If we get to this point, the element is a value, so do nothing!
	return FALSE;
}


// Remove any unary addition operators
BOOL SEquation::SimplifyUnaryPlus(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyUnaryPlus";
#endif // _DEBUG

	if(isBinaryOp(m_ptr->GetType())){
		EqElemBinary *tmp = BINARY(m_ptr);

		BOOL sl = SimplifyUnaryPlus(tmp->L);
		BOOL sr = SimplifyUnaryPlus(tmp->R);

		return sl || sr;
	}

	if(isUnaryOp(m_ptr->GetType())){
		EqElemUnary *tmp = UNARY(m_ptr);

		// Is this element a unary plus?
		if(tmp->GetType() == UNARY_PLUS){

			RemoveUnaryElem(tmp);

			return TRUE;
		}

		return SimplifyUnaryPlus(tmp->C);
	}

	return FALSE;
}


// Remove repeated unary minus signs and move unary minus signs
// towards the root of the tree
BOOL SEquation::SimplifyUnarySub(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyUnarySub";
#endif // _DEBUG

	if(isBinaryOp(m_ptr->GetType())){
		EqElemBinary *tmp = BINARY(m_ptr);

		switch(m_ptr->GetType()){
			case BINARY_DIV:
				if(tmp->L->GetType() == UNARY_NEG){
					// Remove the left child and make a new parent element.
					RemoveUnaryElem(UNARY(tmp->L));
					InsertUnaryParent(tmp, new EqElemUnary(UNARY_NEG));
					return TRUE;
				}

				if(tmp->R->GetType() == UNARY_NEG){
					// Remove the right child and make a new parent element.
					RemoveUnaryElem(UNARY(tmp->R));
					InsertUnaryParent(tmp, new EqElemUnary(UNARY_NEG));
					return TRUE;
				}

				break;
			case BINARY_MULTI:
				if(tmp->L->GetType() == UNARY_NEG){
					// Remove the left child and make a new parent element.
					RemoveUnaryElem(UNARY(tmp->L));
					InsertUnaryParent(tmp, new EqElemUnary(UNARY_NEG));
					return TRUE;
				}

				if(tmp->R->GetType() == UNARY_NEG){
					// Remove the left child and make a new parent element.
					RemoveUnaryElem(UNARY(tmp->R));
					InsertUnaryParent(tmp, new EqElemUnary(UNARY_NEG));
					return TRUE;
				}
				break;
			case BINARY_ADD:
				if(tmp->R->GetType() == UNARY_NEG){
					// A + - B = A - B
					// Turn the current binary operator into a binary subtraction
					// operator and remove the unary negation operator
					tmp->binary_op = BINARY_SUB;

					RemoveUnaryElem(UNARY(tmp->R));

					return TRUE;
				}
				break;
			case BINARY_SUB:
				if(tmp->R->GetType() == UNARY_NEG){
					// A - - B = A + B
					// Turn the current binary operator into a binary addition
					// operator and remove the unary negation operator
					tmp->binary_op = BINARY_ADD;

					RemoveUnaryElem(UNARY(tmp->R));

					return TRUE;
				}
				break;
		}
		
		BOOL sl = SimplifyUnarySub(tmp->L);
		BOOL sr = SimplifyUnarySub(tmp->R);

		return sl || sr;
	}

	if(isUnaryOp(m_ptr->GetType())){
		EqElemUnary *tmp = UNARY(m_ptr);

		// Do we have repeated unary "-" operators?
		if(tmp->GetType() == UNARY_NEG){
			if(tmp->C->GetType() == UNARY_NEG){
				RemoveUnaryElem(UNARY(tmp->C));
				RemoveUnaryElem(UNARY(tmp));

				return TRUE;
			}
		}

		return SimplifyUnarySub(tmp->C);
	}

	return FALSE;
}

// Remove any functions of inverse functions
BOOL SEquation::SimplifyInverse(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyInverse";
#endif // _DEBUG

	if(isBinaryOp(m_ptr->GetType())){
		
		BOOL sl = SimplifyInverse(BINARY(m_ptr)->L);
		BOOL sr = SimplifyInverse(BINARY(m_ptr)->R);

		return sl || sr;
	}

	// Do we have a unary operator acting on a unary operator?
	if(isUnaryOp(m_ptr->GetType()) && isUnaryOp(UNARY(m_ptr)->C->GetType())){
		EqElemUnary *uop = UNARY(m_ptr);
		EqElemUnary *inv_uop = UNARY(UNARY(m_ptr)->C);
		
		// Enumerate all unary functions that have inverses
		switch(uop->GetType()){
			case UNARY_COS:
				// cos(acos(x)) = x
				// Note: acos(cos(x)) != x
				if(inv_uop->GetType() == UNARY_ACOS){
					RemoveUnaryElem(inv_uop);
					RemoveUnaryElem(uop);
					return TRUE;
				}
				break;
			case UNARY_SIN:
				// sin(asin(x)) = x
				// Note: asin(sin(x)) != x
				if(inv_uop->GetType() == UNARY_ASIN){
					RemoveUnaryElem(inv_uop);
					RemoveUnaryElem(uop);
					return TRUE;
				}
				break;
			case UNARY_TAN:
				// tan(atan(x)) = x
				// Note: atan(tan(x)) != x
				if(inv_uop->GetType() == UNARY_ATAN){
					RemoveUnaryElem(inv_uop);
					RemoveUnaryElem(uop);
					return TRUE;
				}
				break;
			case UNARY_EXP:
				// exp(log(x)) = x
				if(inv_uop->GetType() == UNARY_LOG){
					RemoveUnaryElem(inv_uop);
					RemoveUnaryElem(uop);
					return TRUE;
				}
				break;
			case UNARY_LOG:
				// log(exp(x)) = x
				if(inv_uop->GetType() == UNARY_EXP){
					RemoveUnaryElem(inv_uop);
					RemoveUnaryElem(uop);
					return TRUE;
				}
				break;
		}
	}

	if(isUnaryOp(m_ptr->GetType()))
		return SimplifyInverse(UNARY(m_ptr)->C);

	return FALSE;
}


// Remove binary operations ("*/+-") acting on equal, non-constant operands
BOOL SEquation::SimplifyIdentical(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyIdentical";
#endif // _DEBUG
		
	if((m_ptr->GetType() == BINARY_ADD) ||(m_ptr->GetType() == BINARY_SUB)){
		EqElem *tmp_L = BINARY(m_ptr)->L;
		EqElem *tmp_R = BINARY(m_ptr)->R;

		if(FindIdentical(&tmp_L, &tmp_R, m_ptr->GetType())){
#ifdef _DEBUG
			if(tmp_L == NULL)
				throw "tmp_L is NULL in SimplifyIdentical";

			if(tmp_R == NULL)
				throw "tmp_R is NULL in SimplifyIdentical";
#endif // _DEBUG

			// (a + .... - a) --> (a*2 + .... - 0) OR (0 + .... - 0)
			if(GetOp(tmp_L, tmp_R, BINARY(m_ptr)) == BINARY_ADD){
				// (a + .... + a) --> (a*2 + .... - 0)
				// Insert a binary operator
				EqElemBinary *tmp = new EqElemBinary(tmp_L->P);

				if(BINARY(tmp_L->P)->L == tmp_L){
					BINARY(tmp_L->P)->L = tmp;
				}
				else{ // BINARY(tmp_L->P)->R == tmp_L
					BINARY(tmp_L->P)->R = tmp;
				}

				tmp->binary_op = BINARY_MULTI;

				tmp->L = tmp_L;
				tmp->R = MakeConst(2.0, tmp);
				tmp_L->P = tmp;
			}
			else{ // GetOp(tmp_L, tmp_R) == BINARY_SUB
				// (a + .... - a) --> (0 + .... - 0)
				if(BINARY(tmp_L->P)->L == tmp_L){
					BINARY(tmp_L->P)->L = MakeConst(0.0, tmp_L->P);
				}
				else{ // BINARY(tmp_L->P)->R == tmp_L
					BINARY(tmp_L->P)->R = MakeConst(0.0, tmp_L->P);
				}

				tmp_L->CleanUp();
				delete tmp_L;
			}

			if(BINARY(tmp_R->P)->R == tmp_R){
				BINARY(tmp_R->P)->R = MakeConst(0.0, tmp_R->P);
			}
			else{ // BINARY(tmp_R->P)->L == tmp_R
				BINARY(tmp_R->P)->L = MakeConst(0.0, tmp_R->P);
			}

			tmp_R->CleanUp();
			delete tmp_R;

			return TRUE;
		}
	}
	else{
		if((m_ptr->GetType() == BINARY_MULTI) ||(m_ptr->GetType() == BINARY_DIV)){
			EqElem *tmp_L = BINARY(m_ptr)->L;
			EqElem *tmp_R = BINARY(m_ptr)->R;

			if(FindIdentical(&tmp_L, &tmp_R, m_ptr->GetType())){
#ifdef _DEBUG
				if(tmp_L == NULL)
					throw "tmp_L is NULL in SimplifyIdentical";

				if(tmp_R == NULL)
					throw "tmp_R is NULL in SimplifyIdentical";
#endif // _DEBUG

				// (a * .... / a) --> (a^2 * .... / 1) OR (1 * .... / 1)
				if(GetOp(tmp_L, tmp_R, BINARY(m_ptr)) == BINARY_MULTI){

					// (a * .... / a) --> (a^2 * .... / 1)
					// Insert a binary operator
					EqElemBinary *tmp = new EqElemBinary(tmp_L->P);
					
					if(BINARY(tmp_L->P)->L == tmp_L){
						BINARY(tmp_L->P)->L = tmp;
					}
					else{ // BINARY(tmp_L->P)->R == tmp_L
						BINARY(tmp_L->P)->R = tmp;
					}

					tmp->binary_op = BINARY_POW;

					tmp->L = tmp_L;
					tmp->R = MakeConst(2.0, tmp);
					tmp_L->P = tmp;
				}
				else{ // GetOp(tmp_L, tmp_R) == BINARY_DIV
					// (a * .... / a) --> (1 * .... / 1)
					if(BINARY(tmp_L->P)->L == tmp_L){
						BINARY(tmp_L->P)->L = MakeConst(1.0, tmp_L->P);
					}
					else{ // BINARY(tmp_L->P)->R == tmp_L
						BINARY(tmp_L->P)->R = MakeConst(1.0, tmp_L->P);
					}

					tmp_L->CleanUp();
					delete tmp_L;
				}

				if(BINARY(tmp_R->P)->R == tmp_R){
					BINARY(tmp_R->P)->R = MakeConst(1.0, tmp_R->P);
				}
				else{ // BINARY(tmp_R->P)->L == tmp_R
					BINARY(tmp_R->P)->L = MakeConst(1.0, tmp_R->P);
				}

				tmp_R->CleanUp();
				delete tmp_R;

				return TRUE;
			}
		}
	}

	if(isBinaryOp(m_ptr->GetType())){
		BOOL sl = SimplifyIdentical(BINARY(m_ptr)->L);
		BOOL sr = SimplifyIdentical(BINARY(m_ptr)->R);

		return sl || sr;
	}

	if(isUnaryOp(m_ptr->GetType()))
		return SimplifyIdentical(UNARY(m_ptr)->C);

	// Must be value -- no simplification possible

	return FALSE;
}


BOOL SEquation::SimplifyBinarySub(EqElem *m_ptr)
{
#ifdef _DEBUG
	if(m_ptr == NULL)
		throw "NULL pointer in SimplifyBinarySub";
#endif // _DEBUG

	if(m_ptr->GetType() == BINARY_SUB){
		if(BINARY(m_ptr)->R->GetType() == VALUE_CONST){
			EqElemVal *tmp = VALUE(BINARY(m_ptr)->R);

			// Flip the sign
			tmp->constval *= -1.0;

			// Change the binary subtraction into a binary
			// addtion
			BINARY(m_ptr)->binary_op = BINARY_ADD;

			return TRUE;
		}
		
		if(BINARY(m_ptr)->L->GetType() == VALUE_CONST){
			EqElemVal *tmp = VALUE(BINARY(m_ptr)->L);

			// Flip the sign
			tmp->constval *= -1.0;

			// Change the binary subtraction into a binary
			// addtion
			BINARY(m_ptr)->binary_op = BINARY_ADD;
					
			// Insert a unary negation ABOVE the new binary addtion
			InsertUnaryParent(m_ptr, new EqElemUnary(UNARY_NEG));

			return TRUE;
		}
	}

	if(isBinaryOp(m_ptr->GetType())){

		BOOL sl = SimplifyBinarySub(BINARY(m_ptr)->L);
		BOOL sr = SimplifyBinarySub(BINARY(m_ptr)->R);

		return sl || sr;
	}

	if(isUnaryOp(m_ptr->GetType())){
		return SimplifyBinarySub(UNARY(m_ptr)->C);
	}

	return FALSE;
}

