//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
#include "stdafx.h"
#include "saxisobject.h"
#include "splot.h"
#include "Excepts.h"
#include <math.h>

// A fudge factor to
// make sure that the last tick mark
// and tick lable are draw correctly.
#define		TICK_FUDGE_FACTOR	(1.0e-6f)

/////////////////////////////////////////////////////////////////////////////
//	*** SAxisObject ***
/////////////////////////////////////////////////////////////////////////////
SAxisObject::SAxisObject(float m_x, float m_y, float m_line_length, 
						 int m_orientation, float m_tick_min, 
						 float m_tick_max, int m_direction,
						 int m_lable_location)
{
	// Start with a linear scale
	scaleType = SCALE_LINEAR;

	x = m_x;
	y = m_y;
	line_length = m_line_length;
	isLineDrawn = TRUE;
	orientation = m_orientation;

	// The default axis color is black
	axis_color = 0x00000000;

	major_tick_length = 10;
	minor_tick_length = 5;
	
	tick_min = m_tick_min;
	tick_max = m_tick_max;
	
	isStickyTickSpace = FALSE;
	major_tick_space = (tick_max - tick_min)/NUM_MAJOR_TICK;
	minor_tick_space = (tick_max - tick_min)/NUM_MINOR_TICK;

	isTickDrawn = TRUE;
	isGridDrawn = FALSE;

	// Set the default grid color to black
	grid_color = 0;

	grid_size = 1;
	grid_type = SPlot::LINE_SOLID;

	tick_lable_precision = 3;
	direction = m_direction;
	tick_lable_location = m_lable_location;
	isTickLableDrawn = TRUE;
	
	tick_lable_format = GENERAL_LABLE;

	// Set the default tick font color to black
	tick_font_color = 0x00000000;

	// The default font
	tick_font.lfHeight = DefaultTickFontSize;

	tick_font.lfWidth = 0;
	tick_font.lfEscapement = 0;
	tick_font.lfOrientation = tick_font.lfEscapement;
	tick_font.lfWeight = FW_NORMAL;
	tick_font.lfItalic = FALSE;
	tick_font.lfUnderline = FALSE;
	tick_font.lfStrikeOut = FALSE;
	tick_font.lfCharSet = ANSI_CHARSET;
	tick_font.lfOutPrecision	= OUT_DEFAULT_PRECIS;
	tick_font.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	tick_font.lfQuality = DEFAULT_QUALITY;
	tick_font.lfPitchAndFamily = FF_DONTCARE|DEFAULT_PITCH;
	lstrcpy ( tick_font.lfFaceName, DefaultFont );
}

void SAxisObject::PreparePrinting(CDC* pDC, long scaleMul, long scaleDiv, CRect &extent)
{
	SObject::PreparePrinting(pDC, scaleMul, scaleDiv, extent);
	
	tick_font.lfHeight = -MulDiv(-tick_font.lfHeight, printScaleMul, printScaleDiv);
	major_tick_length = MulDiv(major_tick_length, printScaleMul, printScaleDiv);
	minor_tick_length = MulDiv(minor_tick_length, printScaleMul, printScaleDiv);
}

void SAxisObject::CleanUpPrinting(CDC* pDC, CRect &extent)
{	
	tick_font.lfHeight = -MulDiv(-tick_font.lfHeight, printScaleDiv, printScaleMul);
	major_tick_length = MulDiv(major_tick_length, printScaleDiv, printScaleMul);
	minor_tick_length = MulDiv(minor_tick_length, printScaleDiv, printScaleMul);

	SObject::CleanUpPrinting(pDC, extent);
}

void SAxisObject::Draw(CDC* pDC, CRect &extent)
{			
	if(!isDrawn)
		return;

	float major_dt = line_length*major_tick_space/(tick_max - tick_min);
	float minor_dt = line_length*minor_tick_space/(tick_max - tick_min);

	float l2 = 0.5f*line_length;

	// Create a black solid pen
	LOGBRUSH axisBrush;
	axisBrush.lbStyle = BS_SOLID;
	axisBrush.lbColor = axis_color;

	// Use the extented form of the CPen constructor to enable the use
	// of the PS_ENDCAP_SQUARE style. This makes the axis lines pretty when
	// the line scaling is increased.
	CPen pen(PS_GEOMETRIC | PS_ENDCAP_SQUARE, lineScale, &axisBrush);

	CPen *oldpen = pDC->SelectObject(&pen);

	// Draw the axis line
	if(isLineDrawn){
		switch(orientation){
			case HORIZONTAL:
				pDC->MoveTo(XtoScreen(x - l2, extent), YtoScreen(y, extent)); 
				pDC->LineTo(XtoScreen(x + l2, extent), YtoScreen(y, extent)); 
				break;
			case VERTICAL:
				pDC->MoveTo(XtoScreen(x, extent), YtoScreen(y - l2, extent)); 
				pDC->LineTo(XtoScreen(x, extent), YtoScreen(y + l2, extent)); 
				break;
			default:
				throw Error("Illegal axis orientation", 
							__FILE__, __LINE__);
		};
	}

	// Draw the tick marks
	if(isTickDrawn){
		DrawTickMarks(pDC, extent, major_dt, minor_dt, l2);
	}

	// Draw the tick labels
	if(isTickLableDrawn){
		DrawTickLabels(pDC, extent, major_dt, minor_dt, l2);
	}

	// Restore the previous state
	pDC->SelectObject(oldpen);

	if(isSelected && !isPrinting)
		DrawSelect(pDC, extent);
}

void SAxisObject::DrawLabel(CDC* pDC, CRect &extent, float label, float label_loc)
{
	static char format[32];
	static char mantissa[16];
	static char exponent[16];
	int mantissa_len;
	int exponent_len;

	CSize lable_size;
	
	// The numerical value of the exponent
	int n;
	float pow_n;
	float tmp_label;

	if(scaleType == SCALE_LINEAR){

		// The numerical value of the exponent
		n = (int)(floor(log10(fabs(label))));
		pow_n = (float)(pow(10.0, n));

		switch(tick_lable_format){
			case GENERAL_LABLE:
				// No scientific exponent
				exponent_len = 0;

				if((n < -2) || (n > 2)){
					label /= pow_n;

					// Gard against round off error
					sprintf(format, "%%#.%df", tick_lable_precision);
					sprintf(mantissa, format, label);
					if(fabs(atof(mantissa)) >= 10.0){
						n++;

						if((n < -2) || (n > 2)){
							label /= 10.0f;
						}
						else{
							label *= pow_n;

							if(tick_lable_precision == 0)
								// Do not print a trailing decimal point when the
								// precision is set to zero.
								sprintf(format, "%%.%df", tick_lable_precision);
							else
								sprintf(format, "%%#.%df", tick_lable_precision);
							
							sprintf(mantissa, format, label);
							break;
						}
					}

					sprintf(format, "%%#.%dfe%%d", tick_lable_precision);
					sprintf(mantissa, format, label, n);
				}
				else{
					if(tick_lable_precision == 0)
						// Do not print a trailing decimal point when the
						// precision is set to zero.
						sprintf(format, "%%.%df", tick_lable_precision);
					else
						sprintf(format, "%%#.%df", tick_lable_precision);

					sprintf(mantissa, format, label);
				}
				break;
			case ENGINEERING_LABLE:
				// No scientific exponent
				exponent_len = 0;

				if(n != 0)
					label /= pow_n;
				
				// Make sure round off does not lead to a case like 10x10^n
				sprintf(mantissa,"%.*f", tick_lable_precision, label);

				tmp_label = (float)atof(mantissa);
				if(fabs(tmp_label) >= 10.0f){
					label /= 10.0f;
					n++;
				}

				if(tick_lable_precision == 0)
					// Do not print a trailing decimal point when the
					// precision is set to zero.
					sprintf(format, "%%.0fE%%d");
				else
					sprintf(format, "%%#.%dfE%%d", tick_lable_precision);

				sprintf(mantissa, format, label, n);
				break;
			case SCIENTIFIC_LABLE:
				if(n != 0)
					label /= pow_n;
			
				// Make sure round off does not lead to a case like 10x10^n
				sprintf(mantissa,"%.*f", tick_lable_precision, label);

				tmp_label = (float)atof(mantissa);
				if(fabs(tmp_label) >= 10.0f){
					label /= 10.0f;
					n++;
				}

				if(tick_lable_precision == 0)
					// Do not print a trailing decimal point when the
					// precision is set to zero.
					sprintf(format, "%%.0fx10");
				else
					sprintf(format, "%%#.%dfx10", tick_lable_precision);
				
				sprintf(mantissa, format, label);

				sprintf(exponent, "%d", n);
				exponent_len = strlen(exponent);
				break;
		};
	}
	else{ // scaleTytpe == SCALE_LOG
		n = (int)label;
		
		sprintf(mantissa, "10");
		sprintf(exponent, "%d", n);
		exponent_len = strlen(exponent);
	}

	mantissa_len = strlen(mantissa);

	lable_size = pDC->GetTextExtent(mantissa, mantissa_len);
	
	// Store the string location
	int tmp_x, tmp_y;

	if(orientation == HORIZONTAL){
		if(tick_lable_location == UP){						
			pDC->TextOut(
				tmp_x = XtoScreen(label_loc, extent) - lable_size.cx/2, 
				tmp_y = YtoScreen(y, extent) - 2*major_tick_length - lable_size.cy, 
				mantissa, mantissa_len);
		}
		else{ // direction == DOWN
			pDC->TextOut(
				tmp_x = XtoScreen(label_loc, extent) - lable_size.cx/2, 
				tmp_y = YtoScreen(y, extent) + 2*major_tick_length, 
				mantissa, mantissa_len);
		}
	}
	else{ // orientation == VERTICAL
		if(tick_lable_location == LEFT){
			pDC->TextOut(
				tmp_x = XtoScreen(x, extent) - 2*major_tick_length - lable_size.cx, 
				tmp_y = YtoScreen(label_loc, extent) - lable_size.cy/2,
				mantissa, mantissa_len);
		}
		else{ // direction == RIGHT
			pDC->TextOut(
				tmp_x = XtoScreen(x, extent) + 2*major_tick_length, 
				tmp_y = YtoScreen(label_loc, extent) - lable_size.cy/2,
				mantissa, mantissa_len);
		}
	}

	// Do we need to print an exponent?
	if(exponent_len){
		DrawExponent(pDC, exponent, exponent_len, tmp_x + lable_size.cx, tmp_y);
	}
}

void SAxisObject::DrawExponent(CDC* pDC, char text[], int n, int x, int y)
{
	// Change fonts to use superscript
	int orig_height = tick_font.lfHeight;
	
	// Note that font sizes are negative integers!
	if(tick_font.lfHeight < -2)
		tick_font.lfHeight = MulDiv(tick_font.lfHeight, 2, 3);

	CFont super_font;
	super_font.CreateFontIndirect(&tick_font);
	
	// Restore the LOGFONT structure to it original state
	tick_font.lfHeight = orig_height;

	CFont *orig_font = pDC->SelectObject(&super_font);

	pDC->TextOut(x, y, text, n);

	// Restore the font
	pDC->SelectObject(orig_font);
}

void SAxisObject::DrawSelect(CDC* pDC, CRect &extent)
{
	// Draw a dotted selection boxe around the text string
	// Select ROP2 mode
	int ropOld = pDC->SetROP2(R2_XORPEN);

	// Note - since these lines ONLY appear on the screen,
	// they do not have to be scaled by SObject::lineScale!
	CPen linepen(PS_DOT, 1, COLORREF (0) /* Black */);

	// Make the current pen the state pen
	CPen *oldpen = pDC->SelectObject(&linepen); 

	SelectRect(pDC, extent).Draw(pDC);

	// Restore the previous pen
	pDC->SelectObject(oldpen); 
	pDC->SetROP2(ropOld);
}

CSelectRect SAxisObject::SelectRect(CDC* pDC, CRect &extent)
{
	if(!isLineDrawn)
		// No line - so no selection box either.
		return CSelectRect();

	CSelectRect select_rect;

	// This is just a first pass - will need to take into account the
	// tick marks and tick direction.
	select_rect.X(XtoScreen(x, extent));
	select_rect.Y(YtoScreen(y, extent));

	if(orientation == HORIZONTAL){
		select_rect.SizeX(XtoScreen(line_length, extent) - 
			XtoScreen(0.0f, extent));
		select_rect.SizeY(0);
	}
	else{ // orientation == VERTICAL
		select_rect.SizeY(YtoScreen(0.0f, extent) - 
			YtoScreen(line_length, extent));
		select_rect.SizeX(0);
	}

	return select_rect;
}

void SAxisObject::DrawTickLabels(CDC* pDC, CRect &extent, float major_dt, float minor_dt, float l2)
{
	// 't' is an accumulator for measuring distance along an axis. 
	// This variable is stored as a double to prevent round off error from
	// eventually introducing an offset in the tick marks.
	double t;

	float lable;

	CSize label_size;
	CFont tmp_font;
	tmp_font.CreateFontIndirect(&tick_font);

	CFont* save_font = pDC->SelectObject(&tmp_font);
		
	// Set the background drawing mode
	int save_bkg_mode = pDC->SetBkMode(TRANSPARENT);
		
	// Set the font color
	pDC->SetTextColor(tick_font_color);

	switch(orientation){
		case HORIZONTAL:
			if(tick_lable_location == UP){
				for(t = x - l2,lable = tick_min;t <= x + l2 + TICK_FUDGE_FACTOR;
					t += major_dt, lable += major_tick_space){
					
					DrawLabel(pDC, extent, lable, (float)t);
				}
			}
			else{ // direction == DOWN
				for(t = x - l2,lable = tick_min;t <= x + l2 + TICK_FUDGE_FACTOR;
					t += major_dt, lable += major_tick_space){
					DrawLabel(pDC, extent, lable, (float)t);
				}
			}
			break;
		case VERTICAL:
			if(tick_lable_location == LEFT){
				for(t = y - l2,lable = tick_min;t <= y + l2 + TICK_FUDGE_FACTOR;
					t += major_dt, lable += major_tick_space){
					DrawLabel(pDC, extent, lable, (float)t);
				}
			}
			else{ // direction == RIGHT
				for(t = y - l2,lable = tick_min;t <= y + l2 + TICK_FUDGE_FACTOR;
					t += major_dt, lable += major_tick_space){
					
					DrawLabel(pDC, extent, lable, (float)t);
				}
			}
			break;
		default:
			throw Error("Illegal axis orientation", 
						__FILE__, __LINE__);
	};

	pDC->SelectObject(save_font);

	// Restore the background drawing mode
	pDC->SetBkMode(save_bkg_mode);
}

void SAxisObject::DrawTickMarks(CDC* pDC, CRect &extent, float major_dt, float minor_dt, float l2)
{
	int j;

	// 't' is an accumulator for measuring distance along an axis. 
	// This variable is stored as a double to prevent round off error from
	// eventually introducing an offset in the tick marks.
	double t;

	switch(orientation){
		case HORIZONTAL:
			if(direction == UP){
				for(t = x - l2;t <= x + l2 + TICK_FUDGE_FACTOR;t += major_dt){
					pDC->MoveTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent)); 
					pDC->LineTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent) - major_tick_length);
						
					// Draw the minor ticks for the log scale
					if(scaleType == SCALE_LOG){
						for(j = 2;j <= 9;j++){
							float tmp_x_minor = (float)t + (float)(log10(j))*major_dt;
							if(tmp_x_minor < x + l2){
								pDC->MoveTo(XtoScreen(tmp_x_minor, extent), 
									YtoScreen(y, extent)); 
								pDC->LineTo(XtoScreen(tmp_x_minor, extent), 
									YtoScreen(y, extent) - minor_tick_length);
							}
						}
					}
				}
				
				// Draw the minor ticks for the linear scale
				if(scaleType == SCALE_LINEAR)
					for(t = x - l2;t <= x + l2;t += minor_dt){
						pDC->MoveTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent)); 
						pDC->LineTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent) - minor_tick_length);
					}
			}
			else{ // direction == DOWN
				for(t = x - l2;t <= x + l2 + TICK_FUDGE_FACTOR;t += major_dt){
					pDC->MoveTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent)); 
					pDC->LineTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent) + major_tick_length);
						// Draw the minor ticks for the log scale
					if(scaleType == SCALE_LOG){
						for(j = 2;j <= 9;j++){
							float tmp_x_minor = (float)t + (float)(log10(j))*major_dt;
							if(tmp_x_minor < x + l2){
								pDC->MoveTo(XtoScreen(tmp_x_minor, extent), 
									YtoScreen(y, extent)); 
								pDC->LineTo(XtoScreen(tmp_x_minor, extent), 
									YtoScreen(y, extent) + minor_tick_length);
							}
						}
					}
				}
				
				// Draw the minor tick marks for the linear scale
				if(scaleType == SCALE_LINEAR)
					for(t = x - l2;t <= x + l2;t += minor_dt){
						pDC->MoveTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent)); 
						pDC->LineTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent) + minor_tick_length);
				}
			}
			break;
		case VERTICAL:
			if(direction == LEFT){
				for(t = y - l2;t <= y + l2 + TICK_FUDGE_FACTOR;t += major_dt){
						pDC->MoveTo(XtoScreen(x, extent), 
								YtoScreen((float)t, extent)); 
						pDC->LineTo(XtoScreen(x, extent) - major_tick_length, 
								YtoScreen((float)t, extent));
					if(scaleType == SCALE_LOG){
						// Draw the minor ticks
						for(j = 2;j <= 9;j++){
							float tmp_y_minor = (float)t + (float)(log10(j))*major_dt;
							if(tmp_y_minor < y + l2){
								pDC->MoveTo(XtoScreen(x, extent), 
									YtoScreen(tmp_y_minor, extent)); 
								pDC->LineTo(XtoScreen(x, extent) - minor_tick_length, 
									YtoScreen(tmp_y_minor, extent));
							}
						}
					}
				}
				
				// Draw the minor tick marks for the linear scale
				if(scaleType == SCALE_LINEAR)
					for(t = y - l2;t <= y + l2;t += minor_dt){
						pDC->MoveTo(XtoScreen(x, extent), 
								YtoScreen((float)t, extent)); 
						pDC->LineTo(XtoScreen(x, extent) - minor_tick_length, 
								YtoScreen((float)t, extent));
				}
				
			}
			else{ // direction == RIGHT
				for(t = y - l2;t <= y + l2 + TICK_FUDGE_FACTOR;t += major_dt){
					pDC->MoveTo(XtoScreen(x, extent), 
								YtoScreen((float)t, extent)); 
					pDC->LineTo(XtoScreen(x, extent) + major_tick_length, 
								YtoScreen((float)t, extent));
					if(scaleType == SCALE_LOG){
						// Draw the minor ticks
						for(j = 2;j <= 9;j++){
							float tmp_y_minor = (float)t + (float)(log10(j))*
								major_dt;
							if(tmp_y_minor < y + l2){
								pDC->MoveTo(XtoScreen(x, extent), 
									YtoScreen(tmp_y_minor, extent)); 
								pDC->LineTo(XtoScreen(x, extent) + minor_tick_length, 
									YtoScreen(tmp_y_minor, extent));
							}
						}
					}
				}
					
				// Draw the minor tick marks for the linear scale
				if(scaleType == SCALE_LINEAR)
					for(t = y - l2;t <= y + l2;t += minor_dt){
						pDC->MoveTo(XtoScreen(x, extent), 
								YtoScreen((float)t, extent)); 
						pDC->LineTo(XtoScreen(x, extent) + minor_tick_length, 
								YtoScreen((float)t, extent));
					}
			}
			break;
		default:
			throw Error("Illegal axis orientation", 
						__FILE__, __LINE__);
	};
}

DWORD* SAxisObject::GetLineStyle(int lt, int &db_size)
{
	DWORD *buffer = NULL;

	switch(lt){
		case SPlot::LINE_SOLID:
			db_size = 2;
			buffer = new DWORD[db_size];

			buffer[0] = 1;
			buffer[1] = 0;
			break;
		case SPlot::LINE_DOTTED:
			db_size = 2;
			buffer = new DWORD[db_size];
			
			buffer[0] = 4;
			buffer[1] = 4;
			break;
		case SPlot::LINE_DASHED:
			db_size = 2;
			buffer = new DWORD[db_size];

			buffer[0] = 18;
			buffer[1] = 8;
			break;
		case SPlot::LINE_DASHDOT:
			db_size = 4;
			buffer = new DWORD[db_size];

			buffer[0] = 8;
			buffer[1] = 6;
			buffer[2] = 2;
			buffer[3] = 6;
			break;
		case SPlot::LINE_DASHDOTDOT:
			db_size = 6;
			buffer = new DWORD[db_size];

			buffer[0] = 8;
			buffer[1] = 4;
			buffer[2] = 2;
			buffer[3] = 4;
			buffer[4] = 2;
			buffer[5] = 4;
			break;
	};
	
	if(isPrinting && buffer){
		for(int i = 0;i < db_size;i++)
			buffer[i] = MulDiv(buffer[i], printScaleMul, printScaleDiv);
	}

	return buffer;
}

// Map local definitions of line styles to the windows definitions.
int SAxisObject::GetLineStyle(int lt)
{
	switch(lt){
		case SPlot::LINE_SOLID:
			return PS_SOLID;
		case SPlot::LINE_DOTTED:
			return PS_DOT;
		case SPlot::LINE_DASHED:
			return PS_DASH;
		case SPlot::LINE_DASHDOT:
			return PS_DASHDOT;
		case SPlot::LINE_DASHDOTDOT:
			return PS_DASHDOTDOT;
	};
	
	return PS_NULL;
}


void SAxisObject::DrawGrid(CDC* pDC, CRect &extent, float grid_length)
{
	if(!isGridDrawn)
		return;

	DWORD *dashBuffer = NULL;
	CPen *gridPen = NULL;

	BOOL use_user_pen = (grid_type != SPlot::LINE_SOLID) && 
						(grid_size*lineScale != 1) && (!is_16_bit);

	if(use_user_pen){
		// Create a grid_color solid pen
		LOGBRUSH gridBrush;
		gridBrush.lbStyle = BS_SOLID;
		gridBrush.lbColor = grid_color;

		// Use the extented form of the CPen constructor to enable the use
		// of the PS_ENDCAP_SQUARE style. This makes the axis lines pretty when
		// the line scaling is increased.
		int dashSize;
		dashBuffer = GetLineStyle(grid_type, dashSize);

		// Use the extented form of the CPen constructor to enable the use
		// of the thick dashed lines.
		gridPen = new CPen(PS_GEOMETRIC | PS_ENDCAP_SQUARE | PS_USERSTYLE, 
			grid_size*lineScale, &gridBrush, dashSize, dashBuffer);
	}
	else{
		gridPen = new CPen(GetLineStyle(grid_type), grid_size*lineScale, grid_color);
	}

	CPen *oldpen = pDC->SelectObject(gridPen);

	int j;

	// 't' is an accumulator for measuring distance along an axis. 
	// This variable is stored as a double to prevent round off error from
	// eventually introducing an offset in the tick marks.
	double t;

	float major_dt = line_length*major_tick_space/(tick_max - tick_min);
	float minor_dt = line_length*minor_tick_space/(tick_max - tick_min);

	float l2 = 0.5f*line_length;

	switch(orientation){
		case HORIZONTAL:
			for(t = x - l2;t < x + l2 - TICK_FUDGE_FACTOR;t += major_dt){
				
				if(t != x - l2){
					pDC->MoveTo(XtoScreen((float)t, extent), 
								YtoScreen(y, extent)); 
					pDC->LineTo(XtoScreen((float)t, extent), 
								YtoScreen(y + grid_length, extent));
				}

				// Draw the minor ticks for the log scale
				if(scaleType == SCALE_LOG){
					for(j = 2;j <= 9;j++){
						float tmp_x_minor = (float)t + (float)(log10(j))*major_dt;
						if(tmp_x_minor < x + l2){
							pDC->MoveTo(XtoScreen(tmp_x_minor, extent), 
								YtoScreen(y, extent)); 
							pDC->LineTo(XtoScreen(tmp_x_minor, extent), 
								YtoScreen(y + grid_length, extent));
						}
					}
				}
			}
			
			// Draw the minor ticks for the linear scale
			if(scaleType == SCALE_LINEAR)
				for(t = x - l2 + minor_dt;t < x + l2 - TICK_FUDGE_FACTOR;t += minor_dt){

					pDC->MoveTo(XtoScreen((float)t, extent), 
							YtoScreen(y, extent)); 
					pDC->LineTo(XtoScreen((float)t, extent), 
							YtoScreen(y + grid_length, extent));
				}
		break;
	case VERTICAL:
			for(t = y - l2;t < y + l2 - TICK_FUDGE_FACTOR;t += major_dt){

				if(t != y - l2){
					pDC->MoveTo(XtoScreen(x, extent), 
								YtoScreen((float)t, extent)); 
					pDC->LineTo(XtoScreen(x + grid_length, extent), 
								YtoScreen((float)t, extent));
				}

				if(scaleType == SCALE_LOG){
					// Draw the minor ticks
					for(j = 2;j <= 9;j++){
						float tmp_y_minor = (float)t + (float)(log10(j))*
							major_dt;
						if(tmp_y_minor < y + l2){

							pDC->MoveTo(XtoScreen(x, extent), 
									YtoScreen(tmp_y_minor, extent)); 
							pDC->LineTo(XtoScreen(x + grid_length, extent), 
								YtoScreen(tmp_y_minor, extent));
						}
					}
				}
			}
					
			// Draw the minor tick marks for the linear scale
			if(scaleType == SCALE_LINEAR)
				for(t = y - l2 + minor_dt;t < y + l2 - TICK_FUDGE_FACTOR;t += minor_dt){

					pDC->MoveTo(XtoScreen(x, extent), 
							YtoScreen((float)t, extent)); 
					pDC->LineTo(XtoScreen(x + grid_length, extent), 
							YtoScreen((float)t, extent));
				}
		break;
	default:
		throw Error("Illegal axis orientation", 
					__FILE__, __LINE__);
	};

	// Restore the original pen (before deleting the gridPen)
	pDC->SelectObject(oldpen);

	if(dashBuffer)
		delete [] dashBuffer;

	if(gridPen)
		delete gridPen;
}

void SAxisObject::Serialize(CArchive& ar)
{
	SObject::Serialize(ar);
		
	SerializeFont(tick_font, ar);
		
	if(ar.IsStoring()){
		ar << tick_font_color; // <-- Added version 1.06
		ar << scaleType;
		ar << line_length;
		ar << orientation;
		ar << major_tick_length;
		ar << minor_tick_length;
		ar << major_tick_space;
		ar << minor_tick_space;
		ar << direction;
		ar << isTickDrawn;
		ar << isGridDrawn; // <-- Added version 1.18
		ar << grid_color;  // <-- Added version 1.18
		ar << grid_size;   // <-- Added version 1.18
		ar << grid_type;   // <-- Added version 1.18
		ar << isStickyTickSpace; // <-- Added version 1.19
		ar << tick_lable_precision;
		ar << tick_lable_location;
		ar << tick_lable_format;
		ar << tick_min;
		ar << tick_max;
		ar << isTickLableDrawn;
		ar << axis_color;	// <-- Added version 1.21
	}
	else{
		if(version < 1.06f)
			// Set the tick label color to black
			tick_font_color = 0x00000000;
		else
			ar >> tick_font_color;

		ar >> scaleType;
		ar >> line_length;
		ar >> orientation;
		ar >> major_tick_length;
		ar >> minor_tick_length;
		ar >> major_tick_space;
		ar >> minor_tick_space;
		ar >> direction;
		ar >> isTickDrawn;

		if(version < 1.18f){
			// Set the tick label color to black
			isGridDrawn = FALSE;
			grid_color = 0;
			grid_size = 1;
			grid_type = SPlot::LINE_SOLID;
		}
		else{
			ar >> isGridDrawn;
			ar >> grid_color;
			ar >> grid_size;
			ar >> grid_type;
		}
		
		if(version < 1.19f){
			isStickyTickSpace = FALSE;
		}
		else{
			ar >> isStickyTickSpace;
		}

		ar >> tick_lable_precision;
		ar >> tick_lable_location;
		ar >> tick_lable_format;
		ar >> tick_min;
		ar >> tick_max;
		ar >> isTickLableDrawn;
	
		if(version < 1.21){
			axis_color = 0x00000000; //Black
		}
		else{
			ar >> axis_color;
		}
	}
}

